"""
    gestion.views_download, un module pour SLM
    - rendu des pages imprimables (via un format PDF)

    Copyright (C) 2023-2024 Georges Khaznadar <georgesk@debian.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from django.http import FileResponse
from django.template.loader import get_template
from django.core.management import call_command
from django.views.decorators.csrf import ensure_csrf_cookie
from django.db.models import Q
from django.utils.safestring import mark_safe
from django.db.models import Max, Min
from django.utils.translation import gettext_lazy as _

from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
import platform
import json
import traceback

from .decorators import i18n

if platform.system() != "Windows":
    pdfmetrics.registerFont(TTFont('Serif',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-Regular.ttf'))
    pdfmetrics.registerFont(TTFont('SerifBd',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-Bold.ttf'))
    pdfmetrics.registerFont(TTFont('SerifIt',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-Italic.ttf'))
    pdfmetrics.registerFont(TTFont('SerifBI',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-BoldItalic.ttf'))


import trml2pdf
from outils.common import safeIdent

import io
import os
import zipfile
from datetime import date, datetime

from .models import Eleves, Classes, MIN_NO_CAUTION, Caution, Parametres, \
    Courriers, Prets, DATE111, Materiel, Eleves_de_siecle, TypeMateriel, \
    Filieres, Series, Disciplines, Niveaux, NOCLASS, LIMITE_ANNEE_SCOLAIRE

from manuels.settings import BASE_DIR, MEDIA_ROOT

@ensure_csrf_cookie
@i18n
def render_pdf(request,template, context, filename=_("téléchargement.pdf")):
    """
    Rendu sous forme d'un fichier PDF ; template est un modèle au format
    RML que définit Reportlab, context est un dictionnaire pour interpréter
    le modèle.

    Pour le standard RML, voir à https://www.reportlab.com/extra/docs/

    @param request la requête
    @param template le modèle
    @param context dictionnaire pour interpréter le modèle
    @param filename suggestion de nom de fichier pour le téléchargement ;
        "download.pdf" par défaut
    """
    tpl = get_template(template)
    try:
        rendered = tpl.render(context)
        buffer = io.BytesIO(trml2pdf.parseString(rendered))
    except Exception:
        tpl = get_template("gestion/slm_print_error.rml")
        err_lines = traceback.format_exc().split("\n")
        buffer = io.BytesIO(
            trml2pdf.parseString(
                tpl.render({
                    "err_lines": err_lines,
                    "template": template,
                })))
    buffer.seek(0)
    return FileResponse(buffer, as_attachment=True, filename=str(filename))

def manuel_bidon():
    """
    Renvoie un manuel bidon s'il n'y a encore rien dans la base de données.
    Si nécessaire, crée par effet de bord :
    - un type de matériel "Manuel scolaire"
    - une discipline "LETTRES"
    - un niveau "Seconde"
    - une filière "Générale"
    - une série "Toutes séries de la filière"
    """
    type_materiel = TypeMateriel.objects.all()
    if type_materiel:
        tm = type_materiel[0]
    else:
        tm = TypeMateriel(
            libelle_long = _("Manuel scolaire"),
            libelle_court = _("Manu"),
            Tarif_Perdu = ".",
            Tarif_Degradation = ".",
            Choix_Tarif_Degradation = False,
            Appliquer_Demi_Tarif = False,
        )
        tm.save()
    d = Disciplines.objects.all()
    if d:
        discipline = d[0]
    else:
        discipline = Disciplines(
            libelle = _("LETTRES"),
            abrege = _("LET"),
        )
        discipline.save()
    n = Niveaux.objects.all()
    if n:
        niveau = n[0]
    else:
        niveau = Niveaux(
            libelle = _("Seconde"),
            abrege = _("2NDE"),
            rang = "E",
        )
        niveau.save()
    f = Filieres.objects.all()
    if f:
        filiere = f[0]
    else:
        filiere = Filieres(
            nom = _("Générale"),
        )
        filiere.save()
    s = Series.objects.all()
    if s:
        serie = s[0]
    else:
        serie =  Series(
            libelle_court = _("toutes"),
            libelle_long = _("Toutes séries de la filière"),
        )
        serie.save()
        
    return Materiel(
        type_materiel = tm,     # instance de TypeMateriel
        titre = _("C'est un livre bidon"),
        editeur = _("Éditeur"),
        ISBN = _("ISBN"),
        discipline = discipline, # instance de Disciplines
        niveau = niveau,         # instance de Niveaux
        filiere = filiere,       # instance de Filieres
        serie = serie,           # instance de Series
    )

def catalogue_print(request):
    manuels = Materiel.objects.all().order_by("-Annee_Depot_Legal")
    if not manuels:
        manuels = [manuel_bidon()]    
    params, logo = constantes()
    return render_pdf(
        request,
        'gestion/catalogue.rml', {
            "manuels": manuels,
            # Translators : this is a format for strftime
            "date": datetime.now().strftime(str(_("%d %m %Y à %H:%M"))),
            "logo": logo,
        }, filename = _("catalogue.pdf"))

@ensure_csrf_cookie
def saveDB(request):
    """
    Enregistrement de la base de données : un fichier d'export au format
    JSON ou SQLITE zippé est renvoyé, selon le paramètre format
    """
    # on vide la table temporaire  qui correspond à Eleves_de_siecle
    # avant tout enregistrement
    Eleves_de_siecle.objects.all().delete()
    format = request.GET.get("format", "")
    buffer = io.BytesIO()
    zipf = zipfile.ZipFile(buffer, "w")
    if format == "json":
        dump = io.StringIO()
        call_command('dumpdata',format='json',indent=2,stdout=dump)
        dump.seek(0)
        zipf.writestr("db.json", dump.read(),compress_type=zipfile.ZIP_DEFLATED)
        zipf.close()
        filename = "db.json.zip"
    else:
        assert(format == "sqlite")
        dbpath = os.path.join(BASE_DIR, "db.sqlite3")
        zipf.writestr("db.sqlite3", open(dbpath, "rb").read(),
                      compress_type=zipfile.ZIP_DEFLATED)
        zipf.close()
        filename = "db.sqlite-{date}.zip".format(date = datetime.now().strftime(
            "%Y-%m-%d_%H-%M-%S"))
    buffer.seek(0)
    return FileResponse(buffer, as_attachment=True, filename=filename)

def prets_mkPDF(request):
    """
    Crée la fiche de prêt pour un élève. Utilise la paramètre posté "nomprenom"
    """
    nomprenom = request.POST.get("nomprenom")
    eleve =  Eleves.objects.get(NomPrenomEleve = nomprenom)
    # sélectionne les prêts à l'élève qui n'ont pas encore été rendus
    manuels = Prets.objects.filter(
        eleve = eleve
    ).filter(date_retour = DATE111, Date_Declaration_Perte = DATE111).order_by(
        "inventaire__materiel__discipline__abrege",
        "inventaire__id"
    )
    lesprix = [m.inventaire.materiel.Prix for m in manuels]
    messages = Courriers.objects.order_by("N").reverse()
    messages_pret = messages[0].Pret_Recommandation
    total = sum((float(p.replace(",",".")) for p in lesprix if p))
    params, logo = constantes()
    return render_pdf(
        request,
        'gestion/fiche_pret.rml', {
            "eleve": eleve,
            "manuels": manuels,
            "messages": messages_pret.split("\n"),
            "date": datetime.now(),
            "logo": logo,
            "sum": total,
            "params": params,
        },
        filename=_("pret_{}.pdf").format(safeIdent(nomprenom))
    )

def avenant_mkPDF(request):
    """
    Crée un avenant à la fiche de prêt pour un élève.
    Utilise les paramètres postés "nomprenom" et "n" (n vaut "1" ou "2")
    """
    nomprenom = request.POST.get("nomprenom")
    n = request.POST.get("n")
    eleve =  Eleves.objects.get(NomPrenomEleve = nomprenom)
    # sélectionne les prêts à l'élève qui n'ont pas encore été rendus
    manuels = Prets.objects.filter(
        eleve = eleve
    ).filter((Q(date_retour = DATE111) & Q(Date_Declaration_Perte = DATE111)) \
             | Q(date_retour = datetime.now().date())).order_by(
        "inventaire__materiel__discipline__abrege",
        "inventaire__id"
    )
    count_total = len(manuels)
    min_max_date = manuels.aggregate(
            m=Min('date_pret'), M=Max('date_pret'), R=Max('date_retour'))
    date_avenant = max(min_max_date["M"], min_max_date["R"]) if \
        min_max_date["M"] else None
    # les livres qui viennent juste d'être rendus ?
    juste_rendus = Prets.objects.filter(
        eleve = eleve
    ).filter(date_retour = date_avenant).order_by(
        "inventaire__materiel__discipline__abrege",
        "inventaire__id"
    )
    # les manuels qui viennent juste d'être prêtés ?
    manuels = manuels.filter(date_pret = date_avenant)
    lesprix = [m.inventaire.materiel.Prix for m in manuels]
    messages = Courriers.objects.order_by("N").reverse()
    messages_pret = messages[0].Pret_Recommandation
    total = sum((float(p.replace(",",".")) for p in lesprix if p))
    params, logo = constantes()
    return render_pdf(
        request,
        f'gestion/avenant{n}_pret.rml', {
            "eleve": eleve,
            "manuels": manuels,
            "juste_rendus": juste_rendus,
            "count_total": count_total,
            "messages": messages_pret.split("\n"),
            "date": datetime.now(),
            "logo": logo,
            "sum": total,
            "params": params,
        },
        filename=_("avenant{}_{}.pdf").format(n,safeIdent(nomprenom))
    )

def imprime_restitution(request):
    """
    imprime une fiche de restitution,
    étant donné les paramètres postés "nomprenom" et "liste_manuels"
    """
    nomprenom = request.POST.get("nomprenom")
    liste_manuels = json.loads(request.POST.get("liste_manuels"))
    eleve =  Eleves.objects.get(NomPrenomEleve = nomprenom)
    # sélectionne les prêts correspondant à liste_manuels
    manuels = Prets.objects.filter(
        inventaire__id__in = liste_manuels, eleve = eleve,
    ).order_by(
        "inventaire__materiel__discipline__abrege",
        "inventaire__id"
    )
    lesprix = [m.inventaire.materiel.Prix for m in manuels]
    messages = Courriers.objects.order_by("N").reverse()
    messages_rendu = messages[0].Retour_Recommandation
    params, logo = constantes()
    tarifs = [m.tarif for m in manuels]
    rendus = [m for m in manuels if m.isRendu]
    return render_pdf(
        request,
        'gestion/fiche_rendu.rml', {
            "eleve": eleve,
            "manuels": manuels,
            "messages": messages_rendu.split("\n"),
            "date": datetime.now(),
            "logo": logo,
            "params": params,
            "total": sum(tarifs),
            "nb_non_rendu": len(manuels) - len(rendus),
            "lesprix": lesprix,
        },
        filename=_("rendu_{}.pdf").format(safeIdent(nomprenom))
    )

def imprime_memoire(request):
    """
    imprime une fiche de mémoire pour pertes
    étant donné les paramètres postés "nomprenom" et "liste_manuels"
    """
    nomprenom = request.POST.get("nomprenom")
    liste_manuels = json.loads(request.POST.get("liste_manuels"))
    eleve =  Eleves.objects.get(NomPrenomEleve = nomprenom)
    # sélectionne les prêts à l'élève qui n'ont pas encore été rendus
    manuels = Prets.objects.filter(
        inventaire__id__in = liste_manuels
    ).filter(Q(Date_Declaration_Perte__gt = "0001-01-01")).order_by(
        "inventaire__materiel__discipline__abrege",
        "inventaire__id"
    )
    lesprix = [m.inventaire.materiel.Prix for m in manuels]
    messages = Courriers.objects.order_by("N").reverse()
    messages_rendu = messages[0].Memoire
    params, logo = constantes()
    tarifs = [m.tarif for m in manuels]
    rendus = [m for m in manuels if m.isRendu]
    return render_pdf(
        request,
        'gestion/fiche_memoire.rml', {
            "eleve": eleve,
            "manuels": manuels,
            "messages": [mark_safe(l) for l in messages_rendu.split("\n")],
            "date": datetime.now(),
            "logo": logo,
            "params": params,
            "total": sum(tarifs),
            "nb_non_rendu": len(manuels) - len(rendus),
            "lesprix": lesprix,
        },
        filename=_("memoire_{}.pdf").format(safeIdent(nomprenom))
    )

def constantes():
    """
    Récupère les "constantes" de l'établissement à la date courante
    Attention, il faut qu'il y ait un enregistrement dans la table
    Parametres, valide pour l'année scolaire en cours !
    @return un doublet params, logo qui sont une instance de
      Parametres et le chemin absolu vers l'image qu'on y trouve
    """
    annee = int(date.today().strftime("%Y"))
    params = Parametres.objects.filter(AnneeScolaire__contains=str(annee)) \
        or Parametres.objects.filter(AnneeScolaire__contains=str(annee-1))
    if params:
        params = params[0]
    else:
        # on crée des constantes bidon
        params = Parametres(
            NomAcademie = _("Test"),
            NomEts = _("Lycée bidon"),
            CodeRNE = _("0590071g"),
            Commune = _("Nulle part"),
            Titre_ChefEts = _("Proviseur"),
            Civilite_ChefEts = _("Mme"),
            NomChefEts = _("La proviseure"),
            Civilite_AgentComptable = _("M"),
            NomAgentComptable = _("L'intendant"),
            AnneeScolaire = datetime.now().strftime("%Y"),
            Adresse = _("1 rue du lycée"),
            CodePostal = "00000",
            Telephone = "010203040506",
        )
    logo = os.path.join(MEDIA_ROOT, str(params.Logo))
    if not os.path.isfile(logo):
        logo = os.path.join(MEDIA_ROOT, "default", "nologo.png")
    return params, logo

@ensure_csrf_cookie
def cahier_cautions(request):
    """
    Présente le cahier des cautions sous une forme imprimable
    Chaque ligne contient les champs d'une instance d'élève :
    numéro de caution, Nom_de_famille, Prenom, date_inscription_coop
    et laisse de la place pour ajouter des commentaires
    """
    # on ne sélectionne que les élèves dont le numéro de caution
    # a été créé un jour par SLM, pas ceux qui ont été hérités de
    # Gemasco.
    cautions = Caution.objects.filter( numero__gt = MIN_NO_CAUTION )
    if not cautions:
        # on crée une liste avec une seule instance bidon de Caution
        bidon = Caution(
            numero = MIN_NO_CAUTION,
            nom = _("Il n'y a pas encore d'élèves"),
            prenom = _("Il faudrait nourrir la base de données !"),
            date = datetime.now(),
            # Translators : this is a name of a French class at the
            # begin of the high school (students are aged 15)
            classe = _("2d00"),
            commentaire=_("Il faudrait entrer quelques élèves dans la base !"),
        )
        cautions = [bidon]
    premiere_date = min((c.date for c in cautions))
    derniere_date = max((c.date for c in cautions))
    if premiere_date.month >= 7:
        year = premiere_date.year
    else:
        year = premiere_date.year - 1
    debut_cautions = date(year, 7, 1)
    if derniere_date.month < 7:
        year = derniere_date.year
    else:
        year = derniere_date.year + 1
    fin_cautions = date(year, 7, 1)
    cautions_par_annee = {}
    annee_cautions = debut_cautions
    year = annee_cautions.year                   # ex: 2023
    while year < fin_cautions.year:
        s_year = f"{year}/{int(str(year)[2:]) + 1}"  # ex: "2023/24"
        next_year = year + 1
        debut = date(year, 7, 1)
        fin = date(next_year, 7, 1)
        cautions_par_annee[year] = {
            "school_year": s_year,
            "cautions": Caution.objects.filter(
                numero__gt = MIN_NO_CAUTION, date__gte = debut, date__lt = fin)
        }
        year = next_year
    return render_pdf(
        request,
        'gestion/cahier_cautions.rml', {
            "cautions_par_annee": cautions_par_annee,
            "NOCLASS": NOCLASS,
            "colWidths": "1.8cm,5.6cm,3.4cm,2.5cm,4.7cm",
            "vpadding": "0.964mm", # ajustement pour des lignes de 8 mm
        },
        filename=_("cahier_cautions.pdf"),
    )

@ensure_csrf_cookie
def listes_eleves(request):
    """
    Présente les listes d'élèves sous une forme imprimable
    Chaque ligne contient concerne un élève :
    numéro de caution, case à cocher (vide initialement) s'il prend des livres,
    Nom_de_famille, Prenom et laisse de la place pour ajouter des commentaires
    """
    classes = [c for c in Classes.classes_importantes() \
               if c.libelle != NOCLASS]
    listes = {
        c: Eleves.objects.filter(classe = c).order_by('NomPrenomEleve') \
        for c in classes
    }
    if not listes:
        # on renvoie un dictionnaire avec un élève bidon
        bidon = Eleves(
            Nom_de_famille = _("Il n'y a pas encore d'élèves"),
            Prenom = _("Il faudrait nourrir la base de données !"),
        )
        listes = {
            _("2d00"): [bidon]
        }
    return render_pdf(
        request,
        'gestion/listes_eleves.rml', {
            "date": datetime.now().strftime("%d/%m/%Y"),
            "listes": listes,
            "min_livres_emoji": 3, # case cochée pour plus de trois livres
            "colWidths": "1.8cm,1cm,5.6cm,3.4cm,6.2cm",
        },
        filename=_("listes_eleves.pdf"),
    )
    



"""
    gestion.views, un module pour SLM
    - rendu des pages web à l'aide de modèles

    Copyright (C) 2023-2024 Georges Khaznadar <georgesk@debian.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from django.shortcuts import render, redirect, HttpResponse
from django.db.models import Q, Count
from django.views.decorators.csrf import ensure_csrf_cookie
import django.db.models.fields.related
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext
from django.utils.translation import activate, get_language_info

from .views_common import render_common
# from .views_download import *
# from .views_stats import *
# from .odfviews import *
from .models import Eleves, Materiel, LivreSupplement, Jeton, Inventaire, \
    Courriers, Classes, les_modeles, modeles_ordre, ETAT_MATERIEL_CHOIX, \
    Parametres, Prets, Niveaux, NOCLASS, Disciplines, Filieres, Series, \
    TypeMateriel, Commandes, Personnels, Details, sans_accent
from .forms import Eleves_Form, Courriers_Form, donnees_utilisateur_Form, \
    Eleves_ancien_Form, Parametres_Form, Materiel_Form
from .decorators import assure_jeton_valide, i18n
from .utils import svg_histogram, svg_actions_datees, urlToQrcode
from manuels.settings import MEDIA_ROOT, MEDIA_URL

from datetime import date, datetime, timedelta
from io import BytesIO
import pyqrcode
import pathlib
from collections import OrderedDict
import json
import math
from pprint import pformat
import re
import os
from subprocess import run
from lxml import etree
import unidecode

import locale
try:
    locale.setlocale(locale.LC_TIME, "fr_FR.UTF-8")
except locale.Error:
    print("Error: Failed to consider the locale fr_FR.UTF-8")

DATE111 = date.fromisoformat('0001-01-01')

# Create your views here.

@ensure_csrf_cookie
@i18n
def index(request):
    return redirect("/apropos#apropos")

@i18n
def jsi18n(request):
    """
    renvoie un catalogue Javascript pour certaines traductions courtes,
    sans contexte
    """
    translations = {
        'Oui': gettext('Oui'),
        'Non': gettext('Non'),
        'OK': gettext('OK'),
        'Valider': gettext('Valider'),
        'Echappement': gettext('Échappement'),
        'Supprimer': gettext('Supprimer'),
        'ERREUR': gettext('ERREUR'),
        'erreurs': gettext("Des erreurs ont été détectées"),
        'date_erreur': gettext("Date mal formée"),
        'Attention': gettext('Attention'),
        'Bof': gettext('Quelque chose s\'est passé de travers.'),
        'Confirm': gettext("Veuillez confirmer"),
        'Attendre': gettext('Attendre ...'),
        'Neuf': gettext('Neuf'),
        'Bon': gettext('Bon'),
        'Autre': gettext('Autre'),
        "code_inconnu": gettext("code inconnu ?"),
    }
    return HttpResponse(
        "const I18N = " + json.dumps(translations),
        content_type="application/javascript"
    )

@ensure_csrf_cookie
@i18n
def parametres(request):
    p = Parametres.objects.all().order_by(
        '-AnneeScolaire')
    return render_common(request, 'gestion/parametres.html',
                         _("Paramètres"),{
         "aide": "parametres",
         "param": p[0] if p else None,
     })

@ensure_csrf_cookie
@i18n
def apropos(request):
      return render_common(
          request, 'gestion/apropos.html', _("À propos"), pour_tous = True)
    
@ensure_csrf_cookie
@i18n
def assistance(request):
      return render_common(request, 'gestion/assistance.html',
                           _("Assistance"), {
          "aide": "assistance",
      })
    
@ensure_csrf_cookie
@i18n
def achats_inventaire(request):
      return render_common(request, 'gestion/achats_inventaire.html',
                           _("Achats/inventaire"),{
                               "aide": "achats_inventaire",
                           })
    
@ensure_csrf_cookie
@i18n
def prets(request):
    """
    Page de gestion des livres prêtés à un élève ;
    Un paramètre peut être posté, pour pré-remplir le champ de nomprenom
    de l'élève
    """
    nomprenom = request.POST.get("nomprenom","")
    return render_common(request, 'gestion/prets.html',
                         _("Livres prêtés, à ..."),{
        "nomprenom": nomprenom,
        "aide": "prets",
    })

@ensure_csrf_cookie
@i18n
def stats(request):
    """
    Affiche diverses statistiques
    """
    ladate = date.today()
    classes = list(Classes.classes_importantes())
    for c in classes:
        c.membres = len(Eleves.avec_prets_en_cours(classe=c.libelle))
        c.membres_pc = round(100 * c.membres / c.compte)
    manuel_catalogue = Materiel.objects.all().order_by("titre")

    class GlobalLivre:
        """
        données globales concernant des livres d'un même niveau/filière
        """
        def __init__(self):
            self.volume = 0
            self.masse = 0
            self.titres = set()
            
        def add(self, n, manuel):
            """
            ajoute les caractéristiques de n manuels (identiques)
            """
            self.volume += n * manuel.m3
            # densité du papier de l'ordre de 0,9
            self.masse += n * manuel.m3 * 900
            self.titres.add(manuel.titre)

        @property
        def les_titres(self):
            return ", ".join(sorted(list(self.titres)))
        
        def __str__(self):
            return _("""\
Valeurs globales : {volume} m³, {masse} kg, {les_titres}""").\
format(**self.__dict__)
        
    volumes = {}
    total = GlobalLivre()
    lesniveaux = set(m.niveau for m in manuel_catalogue)
    lesfilieres = set(m.filiere for m in manuel_catalogue)
    for niv in lesniveaux:
        for fil in lesfilieres:
            key = f"{niv.abrege} + {fil.nom}"
            manuels = Materiel.objects.filter(niveau = niv, filiere = fil)
            volumes[key] = GlobalLivre()
            for m in manuels:
                n = len(Inventaire.objects.filter(materiel = m))
                volumes[key].add(n, m)
                total.add(n,m)
    membres = Eleves.avec_prets_en_cours()
    livres_par_eleves = [e.prets_en_cours for e in membres]
    histogramme_lpe = svg_histogram(
        livres_par_eleves,
        _("Répartition des nombres de livres pour chaque élève"),
        _("livres"),
        bins = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16],
    )
    graphique_prets = None
    graphique_retours = None
    pr = Prets.objects.all()
    if pr:
        graphique_prets = svg_actions_datees(
            (p.date_pret for p in pr), _("date de prêt"))
        graphique_retours = svg_actions_datees(
            (p.date_retour for p in pr.filter(
                ~Q(date_retour = DATE111))),
            _("date de retour")
        )
    
    return render_common(request,'gestion/stats.html', "Statistiques ...", {
        "ladate": ladate,
        "classes": classes,
        "manuel_catalogue": manuel_catalogue,
        "volumes": volumes,
        "total": total,
        "nb_membres": len(membres),
        "histogramme_lpe": mark_safe(histogramme_lpe),
        "graphique_prets": mark_safe(graphique_prets),
        "graphique_retours": mark_safe(graphique_retours),
    })

@ensure_csrf_cookie
@i18n
def inventaire(request):
    """
    Implémente la page d'inventaire
    """
    niveaux = Niveaux.objects.filter(
        rang__gte = "E", rang__lte = "H").order_by("rang")
    categories = Classes.categories()
    return render_common(request,'gestion/inventaire.html',
                         _("Inventaire"),{
        "aide": "achats_inventaire#inventaire",
        "categories": categories,
        "niveaux": niveaux,
        "classes": [c.libelle for c in Classes.classes_importantes()],
    })

@ensure_csrf_cookie
@i18n
def constantes(request):
    constantes = Parametres.objects.all().order_by("AnneeScolaire")
    messages = Courriers.objects.all().order_by("N")
    p = constantes[len(constantes) - 1] if constantes else None
    classes = {c.libelle: c.compte \
               for c in Classes.classes_importantes()}
    return render_common(
        request, 'gestion/constantes.html',
        _("Constantes de l'établissement"),
        {
            "aide": "parametres#constantes",
            "constantes": constantes,
            "messages": messages,
            "param": p,
            "classes": classes,
            "total": sum(classes.values()),
        })

@ensure_csrf_cookie
@i18n
def constantes_modif(request):
    """
    Permet de voir et d'éditer la description de l'établissement
    """
    numero = request.POST.get("numero","")
    clonage = request.POST.get("clonage","")
    saved = False
    # l'instance à modifier sera définie à deux conditions : que le
    # paramètre numero ne soit pas vide, et que ce numéro identifie bien
    # un enregistrement existant.
    instance = Parametres.objects.filter(pk = int(numero)) if numero \
        else None
    instance = instance[0] if instance else None
    # si le paramètre modif_db n'est pas vide, alors le formulaire
    # est repris depuis les valeurs renseignées dans la page web
    # cependant que la clé primaire viendra toujours de l'instance
    # (que celle-ci soit définie ou non) ; si l'instance est indéfinie
    # ça provoquera la création d'un nouvel enregistrement
    form = Parametres_Form(
        request.POST if request.POST.get("modif_db", "") else None,
        request.FILES if request.POST.get("modif_db", "") else None,
        instance=instance)
    # test de validité avant d'enregistrer
    if form.is_valid():
        form.save()
        saved = True
    # affichage de l'interface web
    return render_common(
        request,'gestion/constantes_modif.html',
        _("Constantes de l'établissement"),
        {
            "numero": numero,
            "clonage": clonage,
            "form": form,
            "saved": saved,
        })
    
@ensure_csrf_cookie
@i18n
def catalogue(request):
    manuels = list(Materiel.objects.all().order_by("-Annee_Depot_Legal", "titre"))
    for m in manuels:
        # on enrichit chaque instance de Manuel de données à afficher en html
        # split_categories forme une liste de catégories
        m.split_categories = m.categories.split(", ")
        # split_options insère des <br> dans un nom d'option et
        # modifie MATHEMATIQUES (mot trop long)
        m.split_option = m.lib_option.replace(".&", ". &").\
            replace(" & ", " & ").replace(" ","<br>").\
            replace("-","<br>-").replace("S.P", "S·P").\
            replace(".", ".<br>").replace("MATHEMATIQUES", "MATHS").\
            replace("<br><br>", "<br>")
    return render_common(
        request, 'gestion/catalogue.html',
        _("Catalogue de la Coopérative"),
        {
            "aide": "achats_inventaire#catalogue",
            "manuels": manuels,
        })

@ensure_csrf_cookie
@i18n
def catalogue_modif(request):
    """
    Permet de voir et d'éditer les entrées du catalogue
    """
    numero = request.POST.get("numero","")
    clonage = request.POST.get("clonage","")
    saved = False
    # l'instance à modifier sera définie à deux conditions : que le
    # paramètre numero ne soit pas vide, et que ce numéro identifie bien
    # un enregistrement existant.
    instance = Materiel.objects.filter(pk = int(numero)) if numero \
        else None
    instance = instance[0] if instance else None
    # si le paramètre modif_db n'est pas vide, alors le formulaire
    # est repris depuis les valeurs renseignées dans la page web
    # cependant que la clé primaire viendra toujours de l'instance
    # (que celle-ci soit définie ou non) ; si l'instance est indéfinie
    # ça provoquera la création d'un nouvel enregistrement
    form = Materiel_Form(
        request.POST if request.POST.get("modif_db", "") else None,
        instance=instance)
    # test de validité avant d'enregistrer
    if form.is_valid():
        form.save()
        saved = True
    # affichage de l'interface web
    categories = json.dumps(Classes.categories())
    options = set([e.Lib_Mat_Enseignee_1 for e in Eleves.objects.all()]) | \
         set([e.Lib_Mat_Enseignee_2 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_3 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_4 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_5 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_6 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_7 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_8 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_9 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_10 for e in Eleves.objects.all()])
    options = sorted([ o for o in options if o])
    return render_common(
        request,'gestion/catalogue_modif.html',
        _("Gestion d'un des items du catalogue"),
        {
            "numero": numero,
            "categories": categories,
            "options": options,
            "clonage": clonage,
            "form": form,
            "saved": saved,
        })
    
@ensure_csrf_cookie
@i18n
def purger_prets(request):
    """
    Permet de purger les prêts quand le livre est revenu
    Considère les paramètres postés date1 et date2 : intervalle de temps
    """
    date1 = request.POST.get("date1", "")
    date2 = request.POST.get("date2", "")
    prets_rendus = Prets.objects.filter(
        date_retour__gt = DATE111)
    if date1 :
        prets_rendus = prets_rendus.filter(
            date_pret__gt = date1)
    if date2 :
        prets_rendus = prets_rendus.filter(
            date_retour__lt = date2)
    return render_common(
        request,'gestion/purger_prets.html',
        _("Purger les prêts anciens"),
        {
            "aide": "parametres#purger",
            "prets_rendus" : prets_rendus,
            "date1" : date1,
            "date2" : date2,
        })

@ensure_csrf_cookie
@i18n
def siecle(request):
    """
    Import d'un fichier siecle, éventuellement POSTé (param fichier_siecle)
    """
    liste_doublons_SLM = {}
    liste_eleves_SLM = Eleves.objects.all()
    liste_doublons_SLM = {
        e.sans_accent: e for e in liste_eleves_SLM \
                if len(e.homonymes(liste_eleves_SLM)) > 1
    }
    return render_common(
        request,'gestion/siecle.html',
        _("Importation depuis la base de données SIÈCLE"),
        {
            "aide": "parametres#siecle",
            "liste_doublons_SLM": liste_doublons_SLM,
        })
    
@ensure_csrf_cookie
@i18n
def eleves(request):
    """
    Affiche la page de gestion des enregistrements des élèves ;
    deux sortes de paramètres peuvent être reçus (postés), certains feront
    voir le formulaire relatif à un élève, d'autres feront voir le
    formulaire relatif à une classe
    """
    numero = request.POST.get("numero","")
    nomprenom = request.POST.get("nomprenom","")
    classe = request.POST.get("choix_classe","")
    saved = False
    if classe:
        eleves = Eleves.objects.filter(
            classe__libelle__iexact = classe).order_by('NomPrenomEleve')
        return render(
            request,'gestion/eleves_de_classe.html',
            {
                "classe": classe,
                "eleves": eleves,
            })
        
    if nomprenom:
        found = Eleves.objects.filter(NomPrenomEleve = nomprenom)
        numero = found[0].id if found else None
    else:
        # l'instance à modifier sera définie à deux conditions : que le
        # paramètre numero ne soit pas vide, et que ce numéro identifie bien
        # un enregistrement existant.
        found = Eleves.objects.filter(id = int(numero))\
            if numero  else None
    instance = found[0] if found else None
    # si le paramètre modif_db n'est pas vide, alors le formulaire
    # est repris depuis les valeurs renseignées dans la page web
    # cependant que la clé primaire viendra toujours de l'instance
    # (que celle-ci soit définie ou non) ; si l'instance est indéfinie
    # ça provoquera la création d'un nouvel enregistrement
    post_data = request.POST if request.POST.get("modif_db", "") else None
    # on choisit le formulaire selon no_caution
    form_class = Eleves_ancien_Form if \
        instance and instance.sorte_de_caution == "ancienne" \
        else Eleves_Form
    form = form_class(post_data, instance=instance)
    # test de validité avant d'enregistrer
    if form.is_valid():
        # avant d'enregistrer on retire le drapeau "a_verifier" qui pourrait
        # rester d'une insription provisoire, ainsi que
        # l'autorisation provisoire d'inscription
        eleve = form.save(commit=False)
        eleve.a_verifier=False
        eleve.autorisation_provisoire_inscription = None
        eleve.save()
        saved = True
        # mise à jour des effectifs des classes
        if found:
            found[0].classe.recompte()
        eleve.classe.recompte()
    # affichage de l'interface web
    result = None
    if nomprenom:
        constantes = Parametres.objects.all().order_by("AnneeScolaire")
        param = constantes[len(constantes) - 1]
        ee = Eleves.objects.filter(NomPrenomEleve = nomprenom)
        besoin_carte = ee[0].peut_recevoir_carte_membre if ee else False
        sans_caution = ee[0].sans_caution if ee else False
        caution_rendue = ee[0].caution_rendue if ee else False
        no_caution = ee[0].no_caution if ee else 0
        result= render(
            request,'gestion/eleve_modif.html',
            {
                "numero": numero,
                "nomprenom": nomprenom,
                "besoin_carte": besoin_carte,
                "sans_caution": sans_caution,
                "caution_rendue": caution_rendue,
                "no_caution": no_caution,
                "form": form,
                "param": param,
                "valide": form["Nom_de_famille"].value() is not None and \
                          form["Prenom"].value()
            })
    else:
        # aucun éléve, aucune classe ne sont demandés
        result= render_common(
            request,'gestion/eleves.html',
            _("Gestion de la base Élèves"),
            {
                "aide": "parametres#base_eleves",
                "saved": saved,
            })
    return result
            
@ensure_csrf_cookie
@i18n
def cartes(request):
    """
    Affiche la page de gestion des cartes de membres ;
    Le POST peut apporter les paramètres
    - liste_classes : une liste de classes selectionnées
    - filtrage: une liste de filtre autorisant des classes sélectionnables
    """
    filtrage = [f for f in request.POST.get("filtrage", "").split(",") if f]
    if len(filtrage) == 0:
        filtrage = ["2D", "1G", "1STMG", "TG", "TMC", "TRH"]
    Qexpr= Q(libelle__startswith=filtrage[0])
    for f in filtrage[1:]:
        Qexpr= Qexpr | Q(libelle__startswith=f)
    classes = Classes.objects.filter(compte__gt=0).filter(Qexpr)
    result= render_common(
        request,'gestion/cartes.html',
        _("Gestion des cartes de membre"),
        {
            "aide": "parametres#cartes_membre",
            "classes": classes,
            "filtrage": ",". join(filtrage),
        })
    return result
            
    
@ensure_csrf_cookie
@i18n
def bon_de_commande(request):
    return render_common(
        request,'gestion/bon_de_commande.html',
        _("Bons de commande"),
        {
        })
    
@ensure_csrf_cookie
@i18n
def tarif(request):
    return render_common(
        request,'gestion/tarif.html',
        _("Définition des tarifs"),
        {
        })
    
@ensure_csrf_cookie
@i18n
def codes_barres(request):
    """
    Gestion de codes-barres pour un manuel du catalogue.
    La clé primaire de ce manuel est dans le paramètre posté "numero"
    """
    numero = int(request.POST.get("numero", 0))
    manuel = None
    if numero:
        # on repère le manuel dans le catalogue
        manuel = Materiel.objects.get(id = numero)
        # on rassemble les manuels de ce titre déjà inventoriés
        items = Inventaire.objects.filter(
            materiel=numero).order_by("Date_Achat_Materiel")
        manuels_inventories = items.values("Date_Achat_Materiel").annotate(
            Count("id"))
        etats = [e[1] for e in ETAT_MATERIEL_CHOIX]
    return render_common(
        request,'gestion/codes_barres.html',
        _("Inventaire des manuels / Impression des étiquettes à codes-barres"),
        {
            "aide": "achats_inventaire#ajout_codes",
            "manuel": manuel,
            "manuels_inventories": manuels_inventories,
            "etats": etats,
        })
    
@ensure_csrf_cookie
@i18n
def codes_barres_print(request):
    """
    Impression de codes-barres pour un manuel du catalogue.
    Si le POST contient un paramètre "nombre", c'est pour créer
    de nouveaux manuels à l'inventaire, auquel cas il y a un autre
    paramètres "etat", clé de l'état (Neuf, Bon, etc.) ;
    s'il contient un paramètre "dejavu-???", c'est qu'on veut revoir
    des codes-barres existants ; dans tous les cas il y a un paramètre
    "id_catalogue".
    Le paramètre "tabou" permet de maintenir une liste de positions
    où il ne faut pas imprimer
    """
    nombre = int(request.POST.get("nombre", 0))
    etat = int(request.POST.get("etat", 0))
    id_catalogue = request.POST.get("id_catalogue","")
    tabou = request.POST.get("tabou","[]")
    date_dejavu = [request.POST[k] for k in request.POST \
              if k.startswith("dejavu-")]
    try:
        date_dejavu = date.fromisoformat(date_dejavu[0])
    except (ValueError, IndexError):
        date_dejavu = ""
    codes = []
    if date_dejavu:
        # on a choisi de revoir des manuels déjà dans l'inventaire
        codes = [m.id for \
                 m in Inventaire.objects.filter(
                     Date_Achat_Materiel = date_dejavu,
                     materiel = id_catalogue)]
    elif nombre:
        # on va créer nombre de manuels nouveaux et les inventorier
        # à la date du jour
        date_dejavu = datetime.now()
        materiel_catalogue = Materiel.objects.get(
            id = id_catalogue)
        for n in range(nombre):
            m = Inventaire(
                Date_Achat_Materiel = date_dejavu,
                materiel = materiel_catalogue,
                NumeroDeSerie = "",
                Perdu = 0,
                Etat_Materiel = int(etat),
                a_imprimer = 1,
            )
            m.save()
            codes.append(m.id)
    manuel = Materiel.objects.get(id = id_catalogue)
    return render_common(
        request,'gestion/codes_barres_print.html',
        _("Impression de codes-barres"),
        {
            "aide": "achats_inventaire#impression_codes",
            "manuel": manuel,
            "nombre": nombre,
            "date": date_dejavu,
            "codes": codes,
            "tabou": tabou,
        })

@ensure_csrf_cookie
@i18n
def cartes_membres_print(request):
    """
    Impression de cartes de membres.
    Le POST contient plusieurs paramètres :
      - id_eleves : une liste d'identifiants d'élèves à qui faire une carte
      - classes_chosen : une liste de classes
      - max_nb    : le nombre max de cartes qu'on peut imprimer sur la
                    première page donnée à l'imprimante, au cas où celle-ci
                    n'a pas toutes les positions imprimables.
    au cas où le POST ne contient qu'un paramètre nomprenom, on reconstitue
    tout de suite id_eleves (pour un seul élève), classes_chosen et
    max_nb
    """
    nomprenom = request.POST.get("nomprenom", "")
    max_nb = int(request.POST.get("max_nb","6")) # 0 c'est "pas de limite"
    if nomprenom :
        classes = []
        id_eleves = [Eleves.objects.get(NomPrenomEleve = nomprenom).id]
    else:
        id_eleves = json.loads(request.POST.get("id_eleves","[]"))
        classes = [ c.strip() for c in
                    json.loads(request.POST.get("classes_chosen","[]"))]
    if max_nb < 2:
        max_nb=2
    if max_nb > 8:
        max_nb=6
    eleves = []
    if classes:
        for c in classes:
            eleves += [e  for e in Eleves.objects.filter(Lib_Structure = c) \
                       if e.peut_recevoir_carte_membre]
    elif nomprenom:
        eleves = list(Eleves.objects.filter(NomPrenomEleve = nomprenom))
    # on prépare une prévisualisation... il s'agit de
    # rectangles de 105 × 74.25 px dans un canevas SVG de 210 × 297 px
    # en 4 lignes de deux cartes.
    rects = []
    nomsprenoms = [e.NomPrenomEleve for e in eleves][:max_nb] if max_nb \
        else [e.NomPrenomEleve for e in eleves][:6]
    # attention, si la feuille à imprimer est incomplète, et que
    # les nombre de cartes est plus petit que le nombre d'emplacements
    # disponibles, on imprime à la fin des cartons détachables !
    # c'est à dire, pour ici, on insère des textes vides
    if max_nb and len(nomsprenoms) < max_nb:
        nomsprenoms = [""] * (max_nb-len(nomsprenoms)) + nomsprenoms
    for i in range(3): # il y a 3 × 2 cartes à imprimer dans une feuille A4
        for j in range(2):
            if max_nb and 2*i+j >= max_nb:
                continue # ne pas chercher à mettre un nomprenom absent !
            rects.append(
                """\
                <rect x="{x}" y="{y}" width="105" height="99"
                style="fill:rgb(120,255,120);stroke-width:1;stroke:rgb(0,0,0)"
                />
                <text x="{x1}" y="{y1}" font-size="5px">
                {nomprenom}
                </text>            
                """.format(
                    x=j*210/2, y=i*297/3,
                    x1=20+j*210/2, y1=40+i*297/3,
                    nomprenom = nomsprenoms[2*i+j]
                )
            )
    # nombres de cartes par page
    cartes=[]
    a_imprimer = len(eleves)
    i = 1
    while a_imprimer:
        if max_nb and i==1:
            n = min(max_nb, a_imprimer)
        else:
            n = min (6, a_imprimer)
        cartes.append({"page": i, "imprime": n})
        a_imprimer -= n
        i+=1
    if len(cartes) > 8:
        cartes = cartes[:4] + [{"page": "...", "imprime": "..."}] + cartes[-4:]
    id_eleves = json.dumps([e.pk for e in eleves])
    return render_common(
        request,'gestion/cartes_membres_print.html',
        _("Impression de cartes de membres"),
        {
            "aide": "parametres#cartes_membre_selection",
            "classes" : json.dumps(classes),
            "nomprenom" : nomprenom,
            "id_eleves": id_eleves,
            "len_eleves": len(eleves),
            "rects": rects,
            "max_nb": max_nb,
            "cartes": cartes,
            "derniere": cartes[-1]["page"],
        })

@i18n
def cartes_membres_print_verso(request):
    """
    Impression de versos de cartes de membres.
    Le POST contient les paramètres :
      - nb : le nombre de pages qu'on veut imprimer, une par défaut
    """
    nb = int(request.POST.get("nb", "1"))
    return render_common(
        request,'gestion/cartes_membres_print_verso.html',
        _("Impression des versos de cartes de membre"),
        {
            "aide": "parametres#cartes_verso",
            "nb": nb,
            "cartes": 6 * nb,
        })

@ensure_csrf_cookie
@i18n
def retours(request):
    """
    Page de retour des livres
    les paramètres du POST sont éventuellement nomprenom et codelivre.
    """
    nomprenom = request.POST.get("nomprenom", "")
    livre = request.POST.get("livre", "")
    return render_common(
        request,'gestion/retours.html',
        _("Gestion des retours de manuels"),
        {
            "aide": "retours",
            "nomprenom": nomprenom,
            "livre": livre,
        })

@ensure_csrf_cookie
@i18n
def sauvegarde(request):

    def sauvegarde_date(path):
        """
        renvoie la date et l'heure de création d'un fichier de sauvegarde
        """
        d = datetime.fromtimestamp(path.stat().st_mtime)
        wd = (d.strftime("%A")+"___________________")[:10]
        return wd + d.strftime("%d/%m/%Y à %H:%M")
        
    p = pathlib.Path(MEDIA_ROOT, "sauvegarde")
    sauvegardes_faites = {
        MEDIA_URL + "sauvegarde/" + path.name: sauvegarde_date(path) \
        for path in p.glob('*.zip')
    }
    return render_common(
        request,'gestion/sauvegarde.html',
        _("Sauvegarder la base de données SLM"),
        {
            "aide": "assistance#sauvegarde",
            "sauvegardes": sorted(
                sauvegardes_faites.items(),
                key = lambda x: x[0],
                reverse = True),
        })
    
@ensure_csrf_cookie
@i18n
def messages_modif(request):
    """
    Permet de voir et d'éditer les messages prédéfinis
    """
    numero = request.POST.get("numero","")
    clonage = request.POST.get("clonage","")
    saved = False
    # l'instance à modifier sera définie à deux conditions : que le
    # paramètre numero ne soit pas vide, et que ce numéro identifie bien
    # un enregistrement existant.
    instance = Courriers.objects.filter(
        N = int(numero)) if numero else None
    instance = instance[0] if instance else None
    # si le paramètre modif_db n'est pas vide, alors le formulaire
    # est repris depuis les valeurs renseignées dans la page web
    # cependant que la clé primaire viendra toujours de l'instance
    # (que celle-ci soit définie ou non) ; si l'instance est indéfinie
    # ça provoquera la création d'un nouvel enregistrement
    form = Courriers_Form(
        request.POST if request.POST.get("modif_db", "") else None,
        instance=instance)
    # test de validité avant d'enregistrer
    if form.is_valid():
        form.save()
        saved = True
    # affichage de l'interface web
    return render_common(
        request,'gestion/messages_modif.html',
        _("Les messages prédéfinis"),
        {
            "numero": numero,
            "clonage": clonage,
            "form": form,
            "saved": saved,
        })
    
@i18n
def donnees_utilisateur(request):
    """
    Modification des données utilisateur
    """
    if request.POST:
        form= donnees_utilisateur_Form(request.POST)
        if form.is_valid():
            request.user.last_name = form.cleaned_data.get("nom")
            request.user.first_name = form.cleaned_data.get("prenom")
            request.user.email = form.cleaned_data.get("email")
            if form.cleaned_data["passe1"]:
                request.user.set_password(form.cleaned_data["passe1"])
            request.user.save()
            return redirect(form.cleaned_data["next"])
    else:
        next = request.GET.get("next")
        form= donnees_utilisateur_Form({
            "nom": request.user.last_name,
            "prenom": request.user.first_name,
            "email": request.user.email,
            "next": next,
        })
    return render_common(
        request,'gestion/donnees_utilisateur.html',
        _("Mise à jour des données personnelles pour {user}").
        format(user = request.user.username),
        {
            "form": form,
        })

@i18n
def importeFromGemaSCO(request):
    """
    Importation de fichiers issus de GemaSCO
    """
    modeles = {}
    for mod in modeles_ordre:
        # création d'un arbre de dépendances
        the_class = eval(mod)
        related = [f.related_model.__name__ for f in the_class._meta.fields \
                   if isinstance(f, django.db.models.fields.related.ForeignKey)]
        modeles[mod] = {
            "the_class": the_class,
            "related": json.dumps(related),
            "csvfile": os.path.basename(les_modeles[mod]["csvfile"]),
        }
    return render_common(
        request,'gestion/importeFromGemaSCO.html',
        _("Importation de données de GemaSCO"),
        {
            "modeles": modeles,
            "names": list(modeles.keys()),
        })

@ensure_csrf_cookie
@i18n
def inscription_provisoire(request):
    """
    Affiche la page pour l'inscription provisoire des élèves ;
    Paramètres de GET ou POST: "code_temporaire"
    Paramètres de POST: tous ceux qui sont gérés par le formulaire
    Eleves_Form
    """
    code_temporaire = request.GET.get("code_temporaire","")
    eleve = None
    if code_temporaire:
        eleves = Eleves.objects.filter(
            autorisation_provisoire_inscription=code_temporaire)
        if eleves:
            eleve = eleves[0]
        else:
            eleve = Eleves(
                Nom_de_famille = "À changer",
                Prenom = "À changer",
                classe = Classes.objects.get(libelle=NOCLASS),
                autorisation_provisoire_inscription=code_temporaire,
            )
            eleve.save()
        form = Eleves_Form(instance=eleve)
        # le code reste valide pour recommencer seulement 30 minutes
        code_valide = datetime.fromisoformat(code_temporaire) > \
            datetime.now() - timedelta(minutes=30)
        return render(request, "gestion/inscription_provisoire.html",{
            "page_pour_tous": True,
            "form": form,
            "code_temporaire": code_temporaire,
            "code_valide": code_valide,
            "saisie": True,
        })
    else:
        code_temporaire = request.POST.get("code_temporaire","")
        fin_validite = datetime.fromisoformat(code_temporaire) + \
            timedelta(minutes=30)
        eleve = Eleves.objects.get(autorisation_provisoire_inscription=code_temporaire)
        form = Eleves_Form(request.POST, instance=eleve)
        ok = form.is_valid()
        if ok:
            eleve = form.save(commit=False)
            eleve.a_verifier = True
            eleve.NomPrenomEleve = eleve.Nom_de_famille + " " + eleve.Prenom
            eleve.save()
        return render(request, "gestion/inscription_provisoire.html",{
            "page_pour_tous": True,
            "saisie": False,
            "code_temporaire": code_temporaire,
            "fin_validite": fin_validite,
            "eleve": eleve,
        })
    
@i18n
def invite_inscription_provisoire(request):
    """
    Crée une autorisation temporaire pour saisir les coordonnées d'un nouvel
    élève, et communique avec l'élève par un lien direct ou par QR-code
    """
    d = datetime.now()
    fin_validite = d + timedelta(minutes=30)
    dernier_autorise = d - timedelta(minutes=30)
    # on annule les champs autorisation_provisoire_inscription des élèves
    # qui ont dépassé la fin de validité, à moins qu'ils aient déjà
    # des livres prêtés
    # on efface d'abord les élèves qu'on n'a pas vérifiés dans les 30 minutes
    eleves_perimes = Eleves.objects.filter(
        a_verifier=True,
        autorisation_provisoire_inscription__lt=dernier_autorise)
    for e in eleves_perimes:
        if not(Prets.objects.filter(eleve = e)):
            # il n'y a pas de livres prêtés pour l'élève e, on l'efface
            # ainsi que sa caution automatique
            c = Cautions.objects.get(numero = e.no_caution)
            c.delete()
            e.delete()
        else:
            # au moins un livre a été prêté, l'élève est considéré comme
            # vérifié
            e.a_verifier=False
            e.autorisation_provisoire_inscription = None
            e.save()
    eleves_ok = Eleves.objects.filter(
        a_verifier=False,
        autorisation_provisoire_inscription__lt=dernier_autorise)
    for e in eleves_ok:
        e.autorisation_provisoire_inscription = None
        e.save()
    # on fait la liste des élèves à vérifier : ils auront un lien de mise à jour
    eleves_a_verifier = Eleves.objects.filter(a_verifier=True)
    code_temporaire = d.isoformat()
    url = request.scheme + "://" + \
        request.META["HTTP_HOST"] + \
        "/inscription_provisoire?code_temporaire=" + str(code_temporaire)
    buffer = BytesIO()
    qr = pyqrcode.create(url)
    qr.svg(buffer, scale=8)
    svg = buffer.getvalue().decode().replace(
        '<?xml version="1.0" encoding="UTF-8"?>', '')
    return render_common(
        request, "gestion/invite_inscription_provisoire.html",
        _("Invitation à l'auto-inscription"),
        {
            "aide": "auto_inscription",
            "url": url,
            "fin_validite": fin_validite,
            "svg": svg,
            "eleves_a_verifier": eleves_a_verifier,
        })

@ensure_csrf_cookie
@i18n
def print_etat_classe(request):
    """
    produit un état des prêts par élèves pour une classe,
    prêt à imprimer depuis le navigateur
    POST contient le paramètre classe
    """
    classe = request.POST.get("classe")
    date = datetime.now().strftime("%Y-%m-%d")
    eleves = Eleves.objects.filter(Lib_Structure = classe).order_by('NomPrenomEleve')
    return render(
        request,
        "gestion/print_etat_classe.html",
        {
            "classe": classe,
            "date": date,
            "eleves": OrderedDict( ((e, ("doit" not in e.doit)) for e in eleves) )
        },
    )
    
@ensure_csrf_cookie
@i18n
def livres_pour(request):
    """
    Présente une page imprimable, pour les livres correspondant à une
    catégorie (2D, 1stmg, 1G, ...) ; POST fournit ce paramètre
    """
    categorie = request.POST.get("categorie")
    livres = Materiel.objects.filter(categories__contains = categorie)
    prix_total = sum((float(liv.Prix.replace(",", ".")) if liv.Prix \
                      else 0.0 for liv in livres))
    return render(
        request,
        "gestion/livres_pour.html",
        {
            "livres": livres,
            "categorie": categorie,
            "prix_total": f'{prix_total:.2f}'.replace(".", ","),
        },
    )
   
    
@ensure_csrf_cookie
@i18n
def cautions(request):
    """
    Sert la page qui permet de gérer les cautions versées par les élèves
    Le POST peut apporter les paramètres nomprenom, qui désigne un élève
    et operation qui peut être "", ou "annule' ou "rembourse"
    """
    nomprenom = request.POST.get("nomprenom", "")
    operation = request.POST.get("operation", "")
    eleve = None
    ok = False
    msg = ""
    if nomprenom:
        eleves = Eleves.objects.filter(NomPrenomEleve = nomprenom)
        if eleves:
            eleve = eleves[0]
            ok = True
        else:
            msg = f"L'élève « {nomprenom} » n'a pas été trouvé dans la base"
    return render_common(
        request, 'gestion/cautions.html', _("Cautions"),
        {
            "aide": "cautions",
            "eleve": eleve,
            "operation": operation,
            "msg": msg,
            "ok": ok,
        }
    )
    
@i18n
def revue_du_stock(request):
    """
    Méthode d'inventaire adaptée à la revue du stock
    """
    return render_common(
        request, 'gestion/revue_du_stock.html', _("Revue du stock"),
        {
            "aide": "revue_du_stock",
        }
    )
    
@i18n
def eleve_scan(request):
     """
    Page qui permet aux élèves de faire leur "shopping" dans la réserve
     """
     return render_common(
        request, 'gestion/eleve_scan.html', _("Recherche de livres"),
         {
             "aide": "eleve_scan",
         }
     )

@i18n
def eleve_invite(request):
    """
    Page qui permet aux élèves de visiter la réserve comme une boutique.
    
    Affiche un QR_code, et l'URL correspondante : tous deux dirigent vers
    une page d'accueil, où un jeton périssable passera dans les cookies,
    pour vérifications utltérieures.

    Elle permet aussi d'afficher le QR-code sur une page imprimable.
    """
    # on crée un jeton de durée une heure et on l'enregistre.
    j = Jeton.create(3600)
    j.save()
    request.session["jeton"] = str(j.id)
    # et on fait un peu de ménage à l'occasion
    Jeton.housekeep()
    baseUrl = f"{request.scheme}://{request.get_host()}/"
    url = baseUrl + f"boutique?jeton={j.id}"
    return render_common(
        request, 'gestion/eleve_invite.html', _("Invitation"),
        {
            "url": url,
            "qrcode": urlToQrcode(url),
            "jeton": j,
            "aide": "eleve_invite",
        }
    )
    
@i18n
def boutique(request):
    """
    Entrée dans la Coop en mode boutique
    On inscrit le jeton dans les cookies, et on propose à l'élève
    d'entrer son nom, puis on lui fait choisir son lot de livres
    à demander.
    """
    jetons = Jeton.objects.filter(id = request.GET.get("jeton"))
    if jetons:
        jeton = jetons[0]
        request.session["jeton"] = str(jeton.id)
    else:
        jeton = None
    return render_common(
        request, 'gestion/boutique.html', _("Boutique de livres"),
        {
            "jeton": jeton,
            "aide": "boutique",
        },
        pour_tous = True,
    )

@i18n
def jeton_perime(request):
    """
    Page qui apparaît quand on visite une page protégée par le décorateur
    @assure_jeton_valide et que le jeton de session est périmé.
    """
    return render_common(
        request, 'gestion/jeton_perime.html', _("Il est trop tard"),
        {
            "jeton": request.session.get("jeton"),
            "aide": "jeton_perime",
        },
        pour_tous = True,
    )
    
@assure_jeton_valide
@ensure_csrf_cookie
@i18n
def moisson(request):
    """
    Moisson de livres, par un élève. Le POST contient une liste de livres
    """
    abreges = request.POST.get("livres")
    abreges = abreges.split(",")
    textes = request.POST.get("txt")
    if textes:
        textes = [t.strip() for t in textes.split(",")]
    else:
        textes = len(abreges) * [""]
    manuels = []
    for a,t in zip(abreges, textes):
        ids = [m.id for m in Materiel.objects.filter(abrege = a)]
        manuels.append({"abrege": a, "ids": ids, "txt": t})
    tous_manuels = []
    tous_abreges = {m.abrege for m in Materiel.objects.all()}
    for a in tous_abreges:
        ids = [m.id for m in Materiel.objects.filter(abrege = a)]
        tous_manuels.append({"abrege": a, "ids": ids})
    return render_common(
        request, 'gestion/moisson.html', _("Chercher les livres"),
        {
            "jeton": Jeton.objects.get(id = request.session.get("jeton")),
            "manuels": manuels,
            "tous_manuels": tous_manuels,
            "aide": "moisson",
        },
        pour_tous = True,
    )

@i18n
def correlation_code(request):
    """
    Affiche du code Python pour entreprendre une analyse de corrélation entre
    les "options", c'est à dire les valeurs importées des champs SIÈCLE
    Lib_Mat_enseignee_x (x = 1 à 5), et les abrégés des livres réellement
    empruntés par une cohorte d'élèves.

    Le but est de construire um moteur qui étant donné une "option",
    prédit si un livre doit être emprunté, ou non. Ce code source Python
    est repris dans le fichier views_stats.py
    """
    options = set([e.Lib_Mat_Enseignee_1 for e in Eleves.objects.all()]) | \
         set([e.Lib_Mat_Enseignee_2 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_3 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_4 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_5 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_6 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_7 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_8 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_9 for e in Eleves.objects.all()]) | \
        set([e.Lib_Mat_Enseignee_10 for e in Eleves.objects.all()])
    options = sorted([ o for o in options if o])
    options = {v:i for i,v in enumerate (options)}
    nb_options = len(options)
    abreges = sorted(list(set(m.abrege for m in Materiel.objects.all() if m.abrege)))
    abreges = {v:i for i,v in enumerate(abreges)}
    nb_abreges = len(abreges)
    donnees_eleves = []
    n = 0
    for e in Eleves.objects.all():
        ## on s'intéresse aux élèves qui ont 3 prêts au moins seulement
        codes_livres_dus = e.dus_codes()
        if len(codes_livres_dus) < 4:
            continue
        n += 1
        ## on ne considère que les 10000 premiers pour le moment
        if n > 10000:
            break
        opt = [0]*nb_options
        for i in (2,3,4,5):
            # on considère les matières enseignées n° 2 à 5
            option = getattr(e, f"Lib_Mat_Enseignee_{i}")
            if option:
                opt[options[option]] = 1
        abg = [0]*nb_abreges
        for code in codes_livres_dus:
            i = abreges[Inventaire.objects.get(id = code).materiel.abrege]
            abg[i] = 1
        donnees_eleves.append((opt,abg))
    # beautification de donnees_eleves
    donnees_formatees = ""
    d = pformat(donnees_eleves, compact = True)
    for line in d.split("\n"):
        donnees_formatees += re.sub(r"(([01], ){26})", r"\1\n   ", line)+"\n"
    return render_common(
        request, 'gestion/correlation_code.html',
        _("Code Python de corrélation"),
        {
            "options": options,
            "abreges": abreges,
            "donnees": donnees_formatees,
            "date": datetime.now(),
            "aide": "correlation_code",
        },
    )  

@i18n
def correlation(request):
    """
    Met en évidence des corrélations entre les options et les manuels
    empruntés par les élèves
    """
    return render_common(
        request, 'gestion/correlation.html',
        _("Corrélation entre options et livres"),
        {
           "aide": "correlation",
        },
    )  
    
@i18n
def qr_boutique(request):
    """
    permet d'imprimer les étiquettes pour les QR-codes de boutique
    """
    # estimation du nombre d'étiquettes, en limitant à 70 cm de hauteur
    # soit deux piles de 35 cm, chaque empilement qui correspond à une
    # référence dans le catalogue
    abreges = {m.abrege: m.id for m in Materiel.objects.all()}
    # attention, plusieurs m.id peuvent avoir un même abrégé, seul le
    # dernier m.id sera conservé ici
    nombre_double_piles = {}
    for a in abreges:
        h=0
        for m in Materiel.objects.filter(abrege = a):
            h += m.epaisseur_mm * len(Inventaire.objects.filter(materiel = m))
        nombre_double_piles[a] = math.ceil(h/700)
    piles = [
        {"abrege": a,
         "code": abreges[a],
         "nombre": nombre_double_piles[a]}
        for a in sorted([ab for ab in abreges],
                        key=lambda x: sans_accent(x).lower())
    ]
    premiere_pile =  None
    livre_exemple =  None
    nombre = 0
    if piles:
        premiere_pile = piles[0]
        lex = Materiel.objects.filter(abrege=premiere_pile["abrege"])
        if lex:
            livre_exemple = lex[0]
            nombre = len(Inventaire.objects.filter(
                materiel__abrege = livre_exemple.abrege))
    return render_common(
        request, 'gestion/qr_boutique.html',
        _("QR-Codes pour le passage en boutique"),
        {
            "piles": piles,
            "premiere_pile": premiere_pile,
            "livre_exemple" : livre_exemple,
            "nombre": nombre,
            "aide": "qr_boutique",
        },
    )  
    
@i18n
def prof_livre_invite(request):
    """
    Page qui permet aux profs de proposer, pendant 96 heures, des
    prêts de livres bien spécifiques à certains élèves
    
    Affiche un QR_code, et l'URL correspondante : tous deux dirigent vers
    une page d'accueil, où un jeton périssable passera dans les cookies,
    pour vérifications utltérieures.

    Elle permet aussi d'afficher le QR-code sur une page imprimable.
    """
    # on crée un jeton de durée une heure et on l'enregistre.
    j = Jeton.create(96*3600)
    j.save()
    request.session["jeton"] = str(j.id)
    # et on fait un peu de ménage à l'occasion
    Jeton.housekeep()
    baseUrl = f"{request.scheme}://{request.get_host()}/"
    url = baseUrl + f"livre_supplement?jeton={j.id}"
    livres_supplement = LivreSupplement.objects.all()
    return render_common(
        request, 'gestion/prof_livre_invite.html', _("Livres à emprunter"),
        {
            "url": url,
            "qrcode": urlToQrcode(url),
            "jeton": j,
            "livres_supplement": livres_supplement,
            "aide": "prof_livre_invite",
        }
    )
    
@assure_jeton_valide
@i18n
def livre_supplement(request):
    return render_common(
        request,
        'gestion/livre_supplement.html',
        "Livre en supplément",
        {
            "aide": "livre_supplement",
        },
        pour_tous = True,
    )

@i18n
def livre_supplement_go(request):
    data = request.POST.get("data")
    commentaire = request.POST.get("commentaire")
    ls = LivreSupplement(data = data, commentaire = commentaire)
    ls.save()
    data = json.loads(data)
    livres = [Materiel.objects.get(id = int(id)) for id in data["livres"]]
    eleves = [Eleves.objects.get(id = int(id)) for id in data["eleves"]]
    baseUrl = f"{request.scheme}://{request.get_host()}/"
    return render_common(
        request,
        'gestion/livre_supplement_go.html',
        _("Livre en supplément"),
        {
            "aide": "livre_supplement_go",
            "data": data,
            "livres": livres,
            "eleves": eleves,
            "baseUrl": baseUrl,
        },
        pour_tous = True,
    )
    
@i18n
def verif_plan(request, filename):
    """
    Vérifie qu'un plan contient des données correctement organisées. Le plan
    est un fichier au format SVG. Dans ce plan, on est censé trouver des
    groupes qui contiennent un rectangle et un texte commençant par "classes:" ;
    ces groupes sont nommés « étagères » dans la mesure où on y entrepose les
    livres destinées à un groupe de classes. D'autre part on y trouve aussi des
    groupes contenant un rectangle, un texte et un cercle ; ces groupes ont
    un attribut "class" qui a la valeur "livres", on les appellera « piles » ;
    ils dénotent le placement d'une ou de plusieurs piles de livres,
    le texte est l'abrégé correspondant aux livres à y trouver, le
    cercle à un « repère lumineux » qui devra être proche de
    l'étiquette matérielle à scanner, dénotant le même abrégé.
    =======
    On vérifiera :
    1) que tous les abrégés du fichier SVG correspondent à des abrégés
       enregistrés dans des instances de models.Materiel ;
    2) que les piles sont bien contenues dans une et une seule
       étagère ;
    3) que le placement d'une pile dans une étagère peut être
       considérée comme cohérente avec un couple niveau/abrégé d'une
       instance de models.Materiel au moins.
    =======
    @param filename un chemin vers un fichier SVG, relatif à
           settings.MEDIA_ROOT, mais sans le suffixe .svg
    """
    svgfiles=[]
    svg=""
    svgfile = MEDIA_ROOT + "/" + filename + ".svg"
    if filename and os.path.exists(svgfile):
        svgfiles=[]
        svg = mark_safe(open(svgfile, encoding="utf-8").read())
    else:
        cp = run(f"find {MEDIA_ROOT} -name '*.svg'", shell=True,
                 encoding="utf-8", capture_output = True)
        svgfiles = [f.replace(MEDIA_ROOT+"/", "") for \
                    f in cp.stdout.strip().split("\n")]

    liste_abreges = sorted(
        list(set([
            m.abrege for m in Materiel.objects.all() if m.abrege.strip()])),
        key = lambda x: unidecode.unidecode(x).lower()
    )
    return render_common(
        request,
        'gestion/verif_plan.html',
        _("Vérification de plan"),
        {
            "aide": "verif_plan",
            "filename": filename + ".svg",
            "liste_abreges": json.dumps(liste_abreges),
            "svg": svg,
            "svgfiles": svgfiles,
        })

@i18n
def une_etiquette(request, numero):
    """
    Permet d'imprimer une étiquette pour un livre, après choix de la
    position pour l'imprimante
    @param numero le numéro d'inventaire du livre
    """
    livre = Inventaire.objects.get(id = int(numero))
    manuel = livre.materiel
    return render_common(
        request,
        'gestion/une_etiquette.html',
        _("Imprimer une étiquette"),
        {
            "aide": "une_etiquette",
            "numero" : numero,
            "manuel": manuel,
            "etiquettes_cols": range(3),
            "etiquettes_ligs": range(8),
        })
    
@i18n
def eleves_plus_n(request, niveau, nombre):
    """
    Donne une liste des élèves d'un niveau qui doivent plus de n livres
    @param niveau seconde, premiere ou terminale
    @param nombre nombre de livres dus au moins (zéro pour au moins un livre)
    """
    debut = "2"
    n=""
    if niveau == "secondes":
        debut = "2"
        n = "seconde"
    elif niveau == "premieres":
        debut = "1"
        n = "première"
    elif niveau == "terminales":
        debut = "T"
        n = "terminale"
    """
    eleves = Eleves.objects.filter(classe__libelle__startswith = debut).\
        order_by("classe__libelle", "sans_accent")
    plus_n = {
        e: len(Prets.faits_a(e)) for e in eleves if \
           len(Prets.faits_a(e)) > int(nombre)
    }
    """
    plus_n = {
        e: len(Prets.faits_a(e)) for e in \
        Eleves.avec_prets_qs(int(nombre)).\
        filter(classe__libelle__startswith = debut).\
        order_by("classe__libelle", "sans_accent")
    }
    return render_common(
        request,
        'gestion/eleves_plus_n.html',
        _("Élèves qui doivent des livres"),
        {
            "aide": "eleves_plus_n",
            "eleves": plus_n,
            "n": n,
            "nombre": nombre,
            "date": date.today(),
        })
        
@i18n
def abreges(request, svgfile, multiplicateur):
    """
    Prépare l'impression d'étiquettes pour les abrégés à placarder
    au niveau des piles de livres
    @param svgfile le nom d'un fichier svg où se trouvent localisés tous les
           abrégés (dans des groupes de classe "livres"), positionnés sur
           des étagères.
    @param multiplicateur le nombre d'étiquettes à imprimer par défaut
           pour chaque groupe de classe "livres"
    """
    multiplicateur = float(multiplicateur)
    svg = etree.parse(os.path.join(MEDIA_ROOT, svgfile))
    etiquettes = svg.xpath('//*[name()="svg"]' +
                           '//*[name()="g"][@class="livres"]' +
                           '/*[name()="text"]')
    tous_abreges = [" ".join(e.xpath('*[name()="tspan"]/text()')) for e in etiquettes]

    abreges = OrderedDict()
    for a in sorted(tous_abreges, key = lambda x: unidecode.unidecode(x).lower()):
        if a in abreges:
            abreges[a] += 1
        else:
            abreges[a] = 1
    for a in abreges:
        abreges[a] *= multiplicateur
        abreges[a] = int(abreges[a])
    return render_common(
        request,
        'gestion/abreges.html',
        _("Imprimer les étiquettes d'abrégés"),
        {
            "aide": "abreges",
            "abreges": abreges,
            "svgfile": os.path.basename(svgfile),
            "multiplicateur": multiplicateur,
        })

def choix_eleves(request):
    """
    Crée des sélections d'élèves pour faire des traitements automatisés
    avec la liste choisie, comme imprimer des étiquettes de courrier, etc.
    """
    return render_common(
        request,
        'gestion/choix_eleves.html',
        _("Actions sur des élèves à choisir"),
        {
            "aide": "choix_eleves",
        })

def action_courrier(request):
    """
    Gestion d'étiquettes d'adresses postales pour des élèves choisis
    Le POST contient un paramètre noms
    """
    noms = request.POST.get("noms")
    #eleves = Eleves.objects.filter(NomPrenomEleve__in = noms)
    return render_common(
        request,
        'gestion/action_courrier.html',
        _("Étiquettes pour des courriers aux élèves"),
        {
            "aide": "action_courrier",
            "noms": noms,
            "lignes": list(range(1, 1+8)),
            "colonnes": list(range(1, 1+3)),
        })

    
def essai_svg(request):
    """
    page expérimentale pour bricoler des graphes de dépendances
    """
    return render_common(
        request,
        'gestion/essai_svg.html',
        _("essai SVG"),
        {
            "aide": "",
        })



"""
    etiquettes.views, un module pour SLM
    - rendu des pages web à l'aide de modèles

    Copyright (C) 2023-2024 Georges Khaznadar <georgesk@debian.org>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import re
import io

import labels
from reportlab.graphics import shapes
from reportlab.graphics.barcode import createBarcodeDrawing
from reportlab.lib.colors import toColor
from reportlab.graphics import renderSVG
from datetime import datetime
    
from .version import this_version
from .models import MIN_NO_CAUTION
from manuels.settings import BASE_DIR

import reportlab.rl_config

from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
import platform
import os.path
reportlab.rl_config.warnOnMissingFontGlyphs = 0

if platform.system() != "Windows":
    pdfmetrics.registerFont(TTFont('Times-Roman',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-Regular.ttf'))
    pdfmetrics.registerFont(TTFont('Times-Bold',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-Bold.ttf'))
    pdfmetrics.registerFont(TTFont('Times-Italic',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-Italic.ttf'))
    pdfmetrics.registerFont(TTFont('Times-BoldItalic',
                                   '/usr/share/fonts/truetype/liberation/LiberationSerif-BoldItalic.ttf'))

    pdfmetrics.registerFont(TTFont('Helvetica',
                                   '/usr/share/fonts/truetype/liberation/LiberationSans-Regular.ttf'))
    pdfmetrics.registerFont(TTFont('Helvetica-Bold',
                                   '/usr/share/fonts/truetype/liberation/LiberationSans-Bold.ttf'))
    pdfmetrics.registerFont(TTFont('Helvetica-Italic',
                                   '/usr/share/fonts/truetype/liberation/LiberationSans-Italic.ttf'))
    pdfmetrics.registerFont(TTFont('Helvetica-BoldItalic',
                                   '/usr/share/fonts/truetype/liberation/LiberationSans-BoldItalic.ttf'))

PADDING = 1
BORDER = False # trace les limites des étiquettes ou pas

def pt2mm(long):
    """
    Convertit une longueur donnée en points vers des millimètres
    """
    return 0.3528 * long

def mm2pt(long):
    """
    Convertit une longueur donnée en millimètres vers des points
    """
    return long / 0.3528

def transform2labelPos(transform, specs):
    """
    Déduit d'une matrice de transformation, la position d'une
    étiquette dans une page.
    @param transform l'attribut d'un élément SVG "g", contenant une étiquette
    @param specs une instance de labels.Specification
    @return un dictionnaire qui donne la ligne et la colonne, numérotées
       à partir de zéro
    """
    pattern = re.compile(r" *matrix\(.*,([.\d]+),([.\d]+)\)")
    m = pattern.match(transform)
    x_mm = pt2mm(float(m.group(1)))
    y_mm = pt2mm(float(m.group(2)))
    col = round((x_mm - float(specs.left_margin)) /
                float(specs.label_width + specs.column_gap))
    row = round((float(
        specs.sheet_height - specs.top_margin - specs.label_height) - y_mm) /
                float(specs.label_height + specs.row_gap))
    return {"row": row, "col": col}

class EtiquetteBase:
    """
    Structure de données commune à toutes les impressions sur des
    planches d'étiquettes, ou microperforées.
    """
    
    """
    @var specs une instance de labels.Specification, ajustée pour le format
    "Avery J8159", vendu aussi sous le nom "A12" autrement.
    Voir https://www.avery.fr/produit/etiquette-pour-timbres-j8159-10
    """
    specs = labels.Specification(
        210, 297,                              # format A4
        3, 8,                                  # 3 colonnes, 9 lignes
        63.5, 33.9,                            # largeur, hauteur de l'étiquette
        corner_radius=2,                       # arrondi
        left_margin=7, right_margin=7,         # marges gauche et droite
        top_margin=12.9,                         # marge en haut
        left_padding=PADDING,                  # espacement intérieur
        right_padding=6.5,                     # coquetterie de GemaSCO ?
        top_padding=PADDING,                   #
        bottom_padding=PADDING,                #
        row_gap=0                              # espace entre lignes
    )

    def __init__(self):
        pass
    
    def adjustXY(self, drawable, x, y):
        """
        Ajuste l'origine d'une instance dessinable de reportlab.
        L'ajustement ne se fait que si l'origine initiale est (0,0)
        @param drawable une instace dessinable de reportlab
        @param x abscisse du début en mm, ou "center"
        @param y ordonnée du début en mm, ou "center"
        """
        x1, y1, x2, y2 = drawable.getBounds()
        w = x2 - x1
        h = y2 - y1
        if x1 == 0 and y1 <= 0:
            x = (self.w - w)/2 if x=="center" else mm2pt(x)
            y = (self.h - h)/2 if y=="center" else mm2pt(y)
            if isinstance(drawable, shapes.Drawing):
                drawable.translate(x, y)
            elif isinstance(drawable, shapes.String):
                drawable.x = x
                drawable.y = y
            elif isinstance(drawable, shapes.Image):
                drawable.x = x
                drawable.y = y
        return

    def addParagraph(self,p, x, y, lineheight, maxLines,
                     fontSize = 10, join = ""):
        """
        Ajoute un paragraphe sur le pylabel
        @param p un paragraphe qui peut contenir des retours à la ligne "/n",
                 et qui sont censées être courtes (tenir sur la lageur de
                 l'étiquette pour chaque ligne)
        @param x abscisse où commence le paragraphe
        @para y ordonnée où commence le paragraphe
        @param lineheight hauteur d'un passage à la ligne
        @param maxLines nombre de lignes à ne pas dépasser
        @param fontSize taille de fonte en pt (10 par défaut)
        @param join (vide par défaut) ; s'il n'est pas vide, c'est une chaîne
                    courte qui peut servir à rassembler des morceaux du
                    paragraphe sur une même ligne tant qu'on ne dépasse pas.
        """
        width = self.specs.label_width - \
            (self.specs.left_padding + self.specs.right_padding)
        lignes = [li.strip() for li in p.split("\n") if li.strip()]
        index = 0
        n_ligne=0
        while index < len(lignes):
            printable = lignes[index]
            s = self.string(printable, fontSize=fontSize)
            x1 ,y1, x2, y2 = s.getBounds()
            if x + pt2mm(x2 - x1) >= width:
                self.adjustXY(s, x, y + n_ligne * lineheight)
                self.pylabel.add(s)
                n_ligne +=1
                if n_ligne > maxLines:
                    print("Error: too many lines")
                    return
                index +=1
                continue
            else:
                accu = printable
                while join and index < len(lignes)-1:
                    accu += join + lignes[index+1]
                    s = self.string(accu, fontSize=fontSize)
                    x1 ,y1, x2, y2 = s.getBounds()
                    if x + pt2mm(x2 - x1) >= width:
                        break
                    else:
                        printable = accu
                        index +=1
                # on a fini d'ajouter des morceaux, il est temps
                # d'écrire sur le label
                s = self.string(printable, fontSize=fontSize)
                self.adjustXY(s, x, y - n_ligne * lineheight)
                self.pylabel.add(s)
                n_ligne +=1
                index += 1
                if n_ligne > maxLines:
                    print("Error: too many lines")
                    return
        return

    def drawBarcode(self, width = 72, height = 6, x = 'center', y = 10):
        """
        Inscrit un code barre dans un pylabel ;
        @param width largeur du code-barre, en mm (47 mm par défaut)
        @param height hauteur du code-barre, en mm (6 mm par défaut)
        @param x abscisse du début ("center" par défaut)
        @param y ordonnée du début (10 mm par défaut)
        """
        bc = createBarcodeDrawing(
            'Code128', value="{code:9d}".format(**self.__dict__),
            width = mm2pt(width), height = mm2pt(height))
        dx = -12 # ce décalage a été réglé expérientalement, au pif
        bc.translate(dx, mm2pt(10))
        self.pylabel.add(bc)
        return

    def string(self, s, fontName="Helvetica", fontSize=6):
        """
        Crée une instance de shapes.String avec x,y = 0,0
        @param s la chaîne à afficher
        @param fontname la police de caractères, Helvetica par défaut
        @param la taille de la police, 6 pt par défaut
        """
        return shapes.String(0, 0, s, fontName=fontName, fontSize=fontSize)

    def image(self, path, width, height):
        """
        Crée une instance de shapes.Image avec x,y = 0,0
        @param path chemin de l'image
        @param width largeur
        @param height hauteur
        """
        if path[0] != "/":
            path = os.path.join(BASE_DIR, path)
        return shapes.Image(0, 0, width, height, path)

    @staticmethod
    def draw_etiquette(label, w, h, instance):
        """
        Fonction de rappel qui crée effectivement le dessin d'une étiquette
        @param label une instance de reportlab.graphics.shapes.Drawing qui
           est gérée automatiquement lors du rappel
        @param w la largeur disponible dans l'objet généré automatiquement
        @param h la hauteur disponible dans l'objet généré automatiquement
        @param instance une instance d'Etiquette
        """
        instance.make_label(label, w, h)
        return

class EtiquetteBoutique(EtiquetteBase):
    """
    Structure de données pour les étiquettes à mettre sur les étagères, afin
    de faciliter le passage des élèves en mode "boutique". Les paramètres
    du constructeur sont :
    @param code ce qu'il faut encoder dans le QR-code à gauche
    @param abrege un mot court pour désigner un manuel
    """
    def __init__(self, code, abrege = "abrégé manquant"):
        self.code = code
        self.abrege = abrege
        self.tempfiles = []
        return

    def qr(self, code, size = 85, x = 0, y = 6):
        """
        inscrit un QR-code dans une image
        @param code le code à inscrire
        @param size largeur et hauteur du QR-code (30 mm par défaut)
        @param x abscisse du début (2mm par défaut)
        @param y ordonnée du début (2mm par défaut)
        """
        import qrcode
        import tempfile
        from qrcode.image.styledpil import StyledPilImage
        from qrcode.image.styles.moduledrawers.pil import RoundedModuleDrawer

        qr = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_H)
        qr.add_data(code)
        img = qr.make_image(
            image_factory=StyledPilImage,
            module_drawer=RoundedModuleDrawer()
        )
        tmp = tempfile.NamedTemporaryFile(prefix="etiq-"+code, suffix=".png")
        img.save(tmp.name, format="PNG")
        self.tempfiles.append(tmp) # so the tempfile lives as long as self
        return shapes.Image(x, y, size, size, tmp.name)

    def drawQR(self):
        img = self.qr(self.code)
        self.pylabel.add(img)
        s = self.string(self.code, fontSize=6)
        self.pylabel.add(s)
        return

    def drawAbrege(self):
        mots = re.split(r"[ -]+", self.abrege)
        p = "\n".join(mots)
        self.addParagraph(
            p, 31, 15, lineheight=6, maxLines=3, fontSize=14, join=" ")
        return

    def make_label(self, label, w, h):
        """
        Enchaîne toutes les étapes de création de l'étiquette, pour
        la bibliothèque pylabels.
        """
        self.pylabel, self.w, self.h = label, w, h
        self.drawQR()
        self.drawAbrege()
        return

class EtiquetteJeanBart(EtiquetteBase):
    """
    Implémentation de la structure de données spécifique aux étiquettes
    pour les manuels de la coopérative du lycée Jean Bart. Les paramètres
    du constructeur sont :
    @param code ce qu'il faut encoder dans le code-barre
    @param titre le titre d'un manuel
    @param abrege un mot court, par exemple "ESP" pour un manuel d'Espagnol
    @param date la date officielle de création de l'étiquette
    @param ets le nom d'établissement (par défaut, "Lycée Jean Bart")
    @param loc la localisation de l'établissement (par défaut, "Dunkerque")
    @param log le nom du logiciel générateur (par défaut, "SLM")
    @param ver la version du logiciel (par défaut, None, ce qui fait
       rechercher dans le sous-module version)
    """
    specs = labels.Specification(
        210, 297,                              # format A4
        3, 8,                                  # 3 colonnes, 9 lignes
        63.5, 33.9,                            # largeur, hauteur de l'étiquette
        corner_radius=2,                       # arrondi
        left_margin=7, right_margin=7,         # marges gauche et droite
        top_margin=12.9,                         # marge en haut
        left_padding=PADDING,                  # espacement intérieur
        right_padding=6.5,                     # coquetterie de GemaSCO ?
        top_padding=PADDING,                   #
        bottom_padding=PADDING,                #
        row_gap=0                              # espace entre lignes
    )

    def __init__(self, code, titre, abrege, adate,
                 ets = "Lycée Jean Bart", loc = "Dunkerque",
                 log = "SLM", ver = None):
        EtiquetteBase.__init__(self)
        self.code = code
        self.titre = titre
        self.abrege = abrege
        self.date = adate
        self.ets = ets
        self.loc = loc
        self.log = log
        if ver:
            self.ver = ver
        else:
            self.ver = this_version
        # propriétés qui seront utilisées pour l'écriture sur un pylabel
        self.pylabel = None # instance de pylabel
        self.w = None       # sa largeur, en points
        self.h = None       # sa hauteur, en points
        return

    def drawEts(self, x='center', y=23):
        """
        Inscrit le nom de l'établissement dans un pylabel ;
        attention, ici l'unité est le point typographique ; 1pt = 0.3528 mm
        @param x l'abscisse de la chaîne à tracer ("center" par défaut)
        @param y l'ordonnée de la chaîne à tracer (23 mm par défaut)
        """
        s = self.string("{ets} - {loc}".format(**self.__dict__), fontSize=6)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        return

    def drawTitle(self, x="center", y = 20):
        """
        Inscrit le titre : date de création à l'inventaire et titre du livre
        @param x l'abscisse de la chaîne à tracer ("center" par défaut)
        @param y l'ordonnée de la chaîne à tracer (20 mm par défaut)
        """
        s = self.string("{date} - {titre}".format(**self.__dict__), fontSize=4)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        return

    def drawAbrege(self, x="center", y=5):
        """
        Inscrit l'abrégé sous le code-barre : abrégé et code
        @param x l'abscisse de la chaîne à tracer ("center" par défaut)
        @param y l'ordonnée de la chaîne à tracer (5 mm par défaut)
        """
        s = self.string("{abrege} - {code}".format(**self.__dict__), fontSize=6)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        ## REMARQUE : on définit ci-dessous un encadrement autour de l'abrégé
        ## mais celui-ci n'est pas centré, comme dans Gemasco à l'origine ;)
        p = shapes.PolyLine(
            [mm2pt(v) for v in [2, 3.5, 56, 3.5, 56, 8.5, 2, 8.5, 2, 3.5]],
            strokeWidth = 0.05,
            strokeColor = toColor("rgb(200,200,200)"),
        )
        self.pylabel.add(p)
        return

    def repeatCode(self):
        """
        Répète deux fois le code numérique sur l'étiquette
        """
        s = self.string("{code}".format(**self.__dict__),
                        fontName="Helvetica-Bold", fontSize=4)
        g = shapes.Group(s)
        g.translate(mm2pt(4), mm2pt(23))
        g.rotate(-90)
        self.pylabel.add(g)
        s = s.copy()
        self.adjustXY(s, 45, 17)
        self.pylabel.add(s)
        return

    def drawSignature(self):
        """
        Donne le nom du logiciel générateur et sa version
        """
        s = self.string(
            "{log} version {ver}".\
            format(**self.__dict__),
            fontSize=4)
        g = shapes.Group(s)
        g.translate(mm2pt(54), mm2pt(23))
        g.rotate(-90)
        self.pylabel.add(g)
        return
    
    def make_label(self,label, w, h):
        """
        Enchaîne toutes les étapes de création de l'étiquette, pour
        la bibliothèque pylabels.
        """
        self.pylabel, self.w, self.h = label, w, h
        self.drawBarcode()
        self.drawEts()
        self.drawTitle()
        self.drawAbrege()
        self.repeatCode()
        self.drawSignature()
        return

class CartesMembre(EtiquetteBase):
    """
    Implémentation de la structure de données spécifique aux cartes de membre
    pour la coopérative du lycée Jean Bart. Les paramètres
    du constructeur sont :
    @param eleve une instance de la classe models.Eleves
    """
    specs = labels.Specification(
        210, 297,                              # format A4
        2, 3,                                  # 2 colonnes, 3 lignes
        105, 99,                               # largeur, hauteur de l'étiquette
        corner_radius=2,                       # arrondi
        left_margin=0, right_margin=0,         # marges gauche et droite
        top_margin=0,                          # marge en haut
        left_padding=8,                        # espacement intérieur
        right_padding=8,                       # 
        top_padding=8,                         #
        bottom_padding=8,                      #
        row_gap=0                              # espace entre lignes
    )

    def __init__(self, eleve):
        EtiquetteBase.__init__(self)
        self.eleve = eleve
        self.pylabel = None # instance de pylabel
        self.w = None       # sa largeur, en points
        self.h = None       # sa hauteur, en points
        return

    def drawEts(self, x='center', y=75):
        """
        Inscrit le nom de l'établissement dans un pylabel ;
        attention, ici l'unité est le point typographique ; 1pt = 0.3528 mm
        @param x l'abscisse de la chaîne à tracer ("center" par défaut)
        @param y l'ordonnée de la chaîne à tracer (75 mm par défaut)
        """
        s = self.string("Lycée Jean Bart -- Dunkerque", fontSize=19)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        return

    def drawTitle(self, x="center", y = 70):
        """
        Inscrit le titre 
        @param x l'abscisse de la chaîne à tracer ("center" par défaut)
        @param y l'ordonnée de la chaîne à tracer (70 mm par défaut)
        """
        classe = self.eleve.Lib_Structure
        if classe.startswith("PAS"):
            classe = ".........."
        s = self.string(
            "Coopérative Scolaire      Classe : {classe}".format(classe=classe),
            fontSize=12)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        return

    def drawMembre(self, x="center", y=60):
        """
        Inscrit les données du membre
        @param x l'abscisse de la chaîne à tracer ("center" par défaut)
        @param y l'ordonnée de la chaîne à tracer (60 mm par défaut)
        """
        no = ".........."
        if self.eleve.no_caution >= MIN_NO_CAUTION:
            no = self.eleve.no_caution
        s = self.string("Membre Actif N° : {no}".format(no=no), fontSize=19)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        return

    def drawNomPrenom(self, x="center", y=50):
        """
        Donne le nom du logiciel générateur et sa version. La taille
        de la fonte sera 20 tant que la longueur de NomPrenomEleve ne
        dépasse pas 20 caractères, et elle diminue après ça.
        @param x l'abscisse de la chaîne à tracer ("center" par défaut)
        @param y l'ordonnée de la chaîne à tracer (50 mm par défaut)
        """
        fontSize = 20
        e = self.eleve
        if len(e.NomPrenomEleve) > 20:
            fontSize = round(20*20/len(e.NomPrenomEleve))
        s = self.string("{e.NomPrenomEleve}".format(e=e),
                        fontSize=fontSize)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        return
    
    def drawDate(self, x="center", y=45):
        """
        affiche la date
        """
        if self.eleve.date_inscription_coop is not None:
            thedate = self.eleve.date_inscription_coop
        else:
            thedate = datetime.now().date()
            self.eleve.date_inscription_coop = thedate
            self.eleve.save()
        s = self.string("Inscription : le {date}".\
            format(date=thedate.strftime("%d/%m/%Y")),
                        fontSize=12)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        return

    def drawSuivi(self, x=5, y=40):
        p = shapes.PolyLine(
            [mm2pt(v) for v in
             [x, y, x+80, y]],
            strokeWidth = 0.2,
            strokeColor = toColor("rgb(0,0,0)"),
        )
        self.pylabel.add(p)
        s = self.string(
            "Livres non rendus : " + \
            ".........................................................",
            fontSize=12)
        self.adjustXY(s, x, y-6)
        self.pylabel.add(s)
        s = self.string(
            ".........................................................................",
            fontSize=12)
        self.adjustXY(s, x, y-12)
        self.pylabel.add(s)
        if self.eleve.sans_caution:
            s = self.string(
                "(s.c.) ..........................",
                fontSize=12)
        else:
            s = self.string(
                "Caution rendue : 70 € ☐ .........",
                fontSize=12)
        self.adjustXY(s, x, y-35)
        self.pylabel.add(s)
        return
    
    def drawSignature(self, x=65, y=20):
        s = self.string(
            "Le trésorier :",
            fontSize=8)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        p = shapes.PolyLine(
            [mm2pt(v) for v in
             [x-1, y+5, x+21, y+5, x+21, y-15, x-1, y-15, x-1, y+5]],
            strokeWidth = 0.1,
            strokeColor = toColor("rgb(0,0,0)"),
        )
        self.pylabel.add(p)
        return

    def make_label(self,label, w, h):
        """
        Enchaîne toutes les étapes de création de l'étiquette, pour
        la bibliothèque pylabels.
        """
        self.pylabel, self.w, self.h = label, w, h
        self.drawEts()
        self.drawTitle()
        self.drawMembre()
        self.drawNomPrenom()
        self.drawDate()
        self.drawSuivi()
        self.drawSignature()
        return

class Suivi(EtiquetteBase):
    """
    Implémentation de la structure de données spécifique aux cartes de membre
    pour la coopérative du lycée Jean Bart. Les paramètres
    du constructeur sont :
    @param eleve une instance de la classe models.Eleves
    """
    specs = labels.Specification(
        297,210,                               # format A4 paysage
        2, 1,                                  # 2 colonnes, 1 ligne
        120, 190,                              # largeur, hauteur de l'étiquette
        corner_radius=2,                       # arrondi
        left_margin=12, right_margin=12,       # marges gauche et droite
        top_margin=0,                          # marge en haut
        left_padding=8,                        # espacement intérieur
        right_padding=8,                       # 
        top_padding=8,                         #
        bottom_padding=8,                      #
        row_gap=0                              # espace entre lignes
    )

    def __init__(self, eleve, annee="0000/00", non_rendus=[]):
        EtiquetteBase.__init__(self)
        self.eleve = eleve
        self.annee = annee
        self.non_rendu = non_rendus
        self.pylabel = None # instance de pylabel
        self.w = None       # sa largeur, en points
        self.h = None       # sa hauteur, en points
        return

    def drawAnnee(self, x="center", y=165):
        """
        Inscrit l'année
        """
        s = self.string("Location {annee}".format(annee=self.annee),
                        fontSize=28, fontName="Helvetica-Bold")
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        p = shapes.PolyLine(
            [mm2pt(v) for v in [0, y-2, 125, y-2]],
            strokeWidth = 0.5,
            strokeColor = toColor("rgb(0,0,0)"),
        )
        self.pylabel.add(p)
        return
    
    def drawEts(self, x=5, y=152):
        """
        Inscrit le nom de l'établissement
        """
        s = self.string("Lycée Jean Bart", fontSize=19,
                        fontName="Helvetica-Bold")
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        s = self.string("COOPÉRATIVE SCOLAIRE", fontSize=11.7)
        self.adjustXY(s, x, y-6)
        self.pylabel.add(s)
        return

    def drawClasseNumero(self, x=65, y=152):
        classe = self.eleve.Lib_Structure
        if classe.startswith("PAS"):
            classe = ".........."
        s = self.string("Classe : {classe}".format(classe=classe), fontSize=11)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        if self.eleve.no_caution >= MIN_NO_CAUTION:
            no = self.eleve.no_caution
            if self.eleve.sans_caution:
                no = f"s.c.({self.eleve.no_caution})"
        else:
            no = ".........."
        s = self.string("N° COOP : {no}".format(no=no),
                        fontSize=11)
        self.adjustXY(s, x, y-6)
        self.pylabel.add(s)
        return

    def drawNomPrenom(self, x=0, y=132):
        s = self.string("NOM :", fontSize=16)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        s = self.string(self.eleve.Nom_de_famille, fontSize=16,
                        fontName="Helvetica-Bold")
        self.adjustXY(s, x+25, y)
        self.pylabel.add(s)
        s = self.string("Prénom :", fontSize=14)
        self.adjustXY(s, x, y-5)
        self.pylabel.add(s)
        s = self.string(self.eleve.Prenom, fontSize=14,
                        fontName="Helvetica-Bold")
        self.adjustXY(s, x+25, y-5)
        self.pylabel.add(s)
        return

    def drawOptions(self, x=0, y=120):
        opt = "Option/Spés : " + \
            (self.eleve.Lib_Mat_Enseignee_1 or "")+ "\n" + \
            (self.eleve.Lib_Mat_Enseignee_2 or "") + "\n" + \
            (self.eleve.Lib_Mat_Enseignee_3 or "") + "\n" + \
            (self.eleve.Lib_Mat_Enseignee_4 or "") + "\n" + \
            (self.eleve.Lib_Mat_Enseignee_5 or "")+ "\n" + \
            (self.eleve.Lib_Mat_Enseignee_6 or "")+ "\n" + \
            (self.eleve.Lib_Mat_Enseignee_7 or "")+ "\n" + \
            (self.eleve.Lib_Mat_Enseignee_8 or "")+ "\n" + \
            (self.eleve.Lib_Mat_Enseignee_9 or "")+ "\n" + \
            (self.eleve.Lib_Mat_Enseignee_10 or "")
        # adresse multi-ligne, 4 mm par ligne, 2 lignes au plus
        # on joint les morcreaux par des points-virgules.
        self.addParagraph(opt, x, y, 4, 2, join=" ; ")
        return

    
    def drawAdresse(self, x=0, y=108):
        if self.eleve.Adresse_Repr_Leg and self.eleve.Adresse_Repr_Leg.split():
            adr = "ADRESSE : " + self.eleve.Adresse_Repr_Leg
        else:
            adr = """\
ADRESSE : .....................................................................
..............................................................................
.............................................................................."""
        # adresse multi-ligne, 4 mm par ligne, 4 lignes au plus
        # on joint les morcreaux par des points-virgules.
        self.addParagraph(adr, x, y, 4, 4, join=" ; ")
        return

    def drawTelephones(self, x = 0, y = 90):
        s = self.string(
            "Tél. Repr. Lég. : {t1} Tél. Autre Repr. : {t2}".format(
                t1 = self.eleve.Tel_Portable_Repr_Leg or "..................",
                t2 = self.eleve.Tel_Portable_Autre_Repr_Leg or "..................",
            ),
            fontSize=10)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        s = self.string(
            "Tél. Élève : {t3}".format(
                t3 = self.eleve.Tel_Portable or ".................."
            ),
            fontSize=10)
        self.adjustXY(s, x, y-4)
        self.pylabel.add(s)
        return

    def drawTitreCompta(self, x = 'center', y = 73):
        s = self.string("Détails comptabilité", fontSize=16,
                        fontName="Helvetica-Bold")
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        p = shapes.PolyLine(
            [mm2pt(v) for v in [0, y-2, 125, y-2]],
            strokeWidth = 0.5,
            strokeColor = toColor("rgb(0,0,0)"),
        )
        self.pylabel.add(p)
        return

    def drawLocation(self, x=0, y= 65):
        s = self.string("Location : 55 €", fontSize=12) 
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        s = self.string(
            "Esp. ☐ ; Chq. ☐ ; HDF N°........................",
            fontSize=12)
        self.adjustXY(s, x + 36, y)
        self.pylabel.add(s)
        s = self.string("Masse : 10 €", fontSize=12)
        self.adjustXY(s, x, y-6)
        self.pylabel.add(s)
        s = self.string("Esp. ☐  Chq. ☐", fontSize=12)
        self.adjustXY(s, x + 36, y-6)
        self.pylabel.add(s)
        s = self.string("Caution : 70 €", fontSize=12)
        self.adjustXY(s, x, y-12)
        self.pylabel.add(s)
        s = self.string("Esp. ☐ ; Chq. ☐", fontSize=12)
        self.adjustXY(s, x + 36, y-12)
        self.pylabel.add(s)
        s = self.string(
            "Chq. banque : .........  Titulaire du compte : ....................",
            fontSize=12)
        self.adjustXY(s, x, y-18)
        self.pylabel.add(s)
        return
    
    def drawObservations(self, x=0, y= 38):
        l1 = "........................................................" + \
            "..............................."
        if self.non_rendu:
            # on commence à marquer la liste des livres non rendus
            ll = len(l1)
            l1 = ",".join((str(nr) for nr in self.non_rendu)) + l1
            l1 = l1[:ll]
            s = self.string("Observations : ", fontSize=12)
            self.adjustXY(s, x, y)
            self.pylabel.add(s)
            img = self.image("gestion/static/img/danger.png", 14, 12)
            self.adjustXY(img, x+27, y)
            self.pylabel.add(img)
            s = self.string(l1, fontSize=5)
            self.adjustXY(s, x+33, y)
            self.pylabel.add(s)
        else:
            s = self.string("Observations : " + l1, fontSize=12)
            self.adjustXY(s, x, y)
            self.pylabel.add(s)
        s = self.string(
            "........................................................" + \
            "........................................................"
                        , fontSize=12)
        self.adjustXY(s, x, y-6)
        self.pylabel.add(s)
        s = self.string(
            "........................................................" + \
            "........................................................"
                        , fontSize=12)
        self.adjustXY(s, x, y-12)
        self.pylabel.add(s)
        return
    
    def drawEnregistrement(self, x='center', y= 10):
        s = self.string("Enregistrement comptable", fontSize=16,
                        fontName="Helvetica-Bold")
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        p = shapes.PolyLine(
            [mm2pt(v) for v in [0, y-2, 125, y-2]],
            strokeWidth = 0.5,
            strokeColor = toColor("rgb(0,0,0)"),
        )
        self.pylabel.add(p)
        s = self.string("Livre N° : ....                      Page : ....."
                        , fontSize=12)
        self.adjustXY(s, x, y-8)
        self.pylabel.add(s)
       
        return
    
    def make_label(self,label, w, h):
        """
        Enchaîne toutes les étapes de création de l'étiquette, pour
        la bibliothèque pylabels.
        """
        self.pylabel, self.w, self.h = label, w, h
        self.drawAnnee()
        self.drawEts()
        self.drawClasseNumero()
        self.drawNomPrenom()
        self.drawOptions()
        self.drawAdresse()
        self.drawTelephones()
        self.drawTitreCompta()
        self.drawLocation()
        self.drawObservations()
        self.drawEnregistrement()
        return

def previewSVG(data, tabou):
    """
    Produit un fichier SVG de planches d'étiquettes ; ce fichier est
    enrichi par rapport à ce que donne Reportlab afin d'attribuer un
    propriété d'interaction à chaque étiquette
    @param data une liste d'instances d'EtiquetteJeanBart
    @param tabou une liste de positions où on ne doit pas imprimer
       d'étiquette ; chaque position est donnée par un dictionnaire
       "row" : entier, "col": entier (numéros débutant à un)
    @return un triplet :
       un rendu de la première page au format SVG,
       le nombre d'étiquettes et de pages
    """
    sheet = labels.Sheet(
        EtiquetteJeanBart.specs, EtiquetteJeanBart.draw_etiquette, border=True)
    sheet.partial_page(1,[(int(pos["row"]), int(pos["col"])) for pos in tabou])
    for etiquette in data:
        sheet.add_label(etiquette)
    svg = renderSVG.drawToString(sheet._pages[0])
    # on dégraisse le document pour ne garder que le nœud svg
    from lxml import etree as ET
    def istag(tagname, elt):
        """
        Vérifie un tag indépendemment de l'espace de nom pour lxml
        """
        return re.sub(r"\{.*\}", "", elt.tag) == tagname
    def firstTag(tagname, elt):
        """
        Trouve le premier descendant possédant un tag parmi les
        fils d'un élément
        """
        found = [e for e in elt if istag(tagname, e)]
        return found[0] if found else None
    def children(tagname, elt):
        """
        Renvoie les éléments descendant de elt qui on un tag donné
        """
        return [e for e in elt if istag(tagname, e)]
    
    doc = ET.fromstring(svg.encode())
    title = firstTag("title", doc)
    title.text = "Cliquer sur les étiquettes pour activer/désactiver l'impression"
    # On "descend de deux groupes" dans le document
    # et à ce niveau, il y a une collection de groupes, chacun avec
    # une étiquette.
    etiquettes = children("g",firstTag("g", firstTag("g", doc)))
    for e in etiquettes:
        pos = transform2labelPos(e.attrib["transform"], EtiquetteJeanBart.specs)
        row = pos["row"]
        col = pos["col"]
        # rend l'objet cliquable
        e.attrib["onclick"] = "toggleLabel(this);"
        e.attrib["style"] = "cursor: pointer"
        rect = ET.Element("rect")
        rect.attrib["fill"] = "rgba(200,255,200,0.3)"
        rect.attrib["class"] = "ok"
        rect.attrib["title"] = "Cliquer pour désactiver cette position"
        rect.attrib["width"] = "{}".format(
            mm2pt(float(EtiquetteJeanBart.specs.label_width)))
        rect.attrib["height"] = "{}".format(
            mm2pt(float(EtiquetteJeanBart.specs.label_height)))
        # ajout de données ligne, colonne
        rect.attrib["data-col"] = str(col+1)
        rect.attrib["data-row"] = str(row+1)
        e.append(rect)
    return ET.tostring(doc).decode(), sheet.label_count, sheet.page_count

class CartesMembre_verso(EtiquetteBase):
    """
    Implémentation de la structure de données des versos de cartes de membre
    pour la coopérative du lycée Jean Bart.
    """
    specs = labels.Specification(
        210, 297,                              # format A4
        2, 3,                                  # 2 colonnes, 3 lignes
        105, 99,                               # largeur, hauteur de l'étiquette
        corner_radius=2,                       # arrondi
        left_margin=0, right_margin=0,         # marges gauche et droite
        top_margin=0,                          # marge en haut
        left_padding=8,                        # espacement intérieur
        right_padding=8,                       # 
        top_padding=8,                         #
        bottom_padding=8,                      #
        row_gap=0                              # espace entre lignes
    )

    def __init__(self):
        EtiquetteBase.__init__(self)
        return

    def drawRendu(self, x=5, y=75):
        """
        affiche la ligne de rendu
        """
        s = self.string(
            "Classe : ..........  Caution rendue le : ..../..../....",
            fontSize=12)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        s = self.string(
            "CNI ☐ C. Bus ☐ Badge ☐ Autre ☐",
            fontSize=12)
        self.adjustXY(s, x, y-6)
        self.pylabel.add(s)
        s = self.string(
            "Chèque au guichet ☐ Chèque par courrier ☐",
            fontSize=12)
        self.adjustXY(s, x, y-12)
        self.pylabel.add(s)
        s = self.string(
            "Ordre : ........................................................",
            fontSize=12)
        self.adjustXY(s, x, y-18)
        self.pylabel.add(s)
        return

    def drawCaution(self, x=5, y=40):
        s = self.string(
            "Cette carte vaut pour une caution de 70 €, remboursable",
            fontSize=8)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        s = self.string(
            "dans les 3 mois suivant la sortie de la coopérative.",
            fontSize=8)
        self.adjustXY(s, x, y-4)
        self.pylabel.add(s)
        return

    def drawSignature(self, x=5, y=30):
        s = self.string(
            "Signature :",
            fontSize=8)
        self.adjustXY(s, x, y)
        self.pylabel.add(s)
        p = shapes.PolyLine(
            [mm2pt(v) for v in
             [x-1, y+5, x+80, y+5, x+80, y-15, x-1, y-15, x-1, y+5]],
            strokeWidth = 0.1,
            strokeColor = toColor("rgb(0,0,0)"),
        )
        self.pylabel.add(p)
        return
        
    
    def make_label(self,label, w, h):
        """
        Enchaîne toutes les étapes de création de l'étiquette, pour
        la bibliothèque pylabels.
        """
        self.pylabel, self.w, self.h = label, w, h
        self.drawRendu()
        self.drawCaution()
        self.drawSignature()
        return
    
class EtiquetteAdresse(EtiquetteBase):
    """
    Structure de données pour les étiquettes d'adresse concernant un élève :
    @param e une instance d'Eleves
    """
    def __init__(self, e):
        self.eleve = e
        return

    def drawNom(self):
        self.addParagraph(
            self.eleve.NomPrenomEleve, 2, 25, lineheight=6,
            maxLines=1, fontSize=10, join=" ")
        return

    def drawAdresse(self):
        debut, code = self.eleve.adresse_codepostal()
        self.addParagraph(
            debut, 2, 21, lineheight=4,
            maxLines=5, fontSize=7.5, join=" ")
        self.addParagraph(
            code, 2, 8, lineheight=4,
            maxLines=2, fontSize=10, join=" ")
        return
    
    def make_label(self, label, w, h):
        """
        Enchaîne toutes les étapes de création de l'étiquette, pour
        la bibliothèque pylabels.
        """
        self.pylabel, self.w, self.h = label, w, h
        self.drawNom()
        self.drawAdresse()
        return

def makePDF(data, tabou):
    """
    Produit un fichier PDF de planches d'étiquettes
    @param data une liste d'instances d'EtiquetteJeanBart
    @param tabou une liste de positions où on ne doit pas imprimer
       d'étiquette ; chaque position est donnée par un dictionnaire
       "row" : entier, "col": entier (numéros débutant à un)
    @return un pseudo-fichier, ouvert en lecteur, initialisé à son début
    """
    sheet = labels.Sheet(
        EtiquetteJeanBart.specs, EtiquetteJeanBart.draw_etiquette,
        border=BORDER)
    sheet.partial_page(1,[(int(pos["row"]), int(pos["col"])) for pos in tabou])
    for etiquette in data:
        sheet.add_label(etiquette)
    buffer = io.BytesIO()
    sheet.save(buffer)
    buffer.seek(0)
    return buffer

def make_carte_membre_PDF(data, tabou):
    """
    Produit un fichier PDF de cartes de membre
    @param data une liste d'instances de CartesMembre
    @param tabou une liste de positions où on ne doit pas imprimer
       d'étiquette ; chaque position est donnée par un dictionnaire
       "row" : entier, "col": entier (numéros débutant à un)
    @return un pseudo-fichier, ouvert en lecteur, initialisé à son début
    """
    sheet = labels.Sheet(CartesMembre.specs, CartesMembre.draw_etiquette,
                         border=False)
    sheet.partial_page(1,[(int(pos["row"]), int(pos["col"])) for pos in tabou])
    for etiquette in data:
        sheet.add_label(etiquette)
    buffer = io.BytesIO()
    sheet.save(buffer)
    buffer.seek(0)
    return buffer

def make_suivi_PDF(data):
    """
    Produit un fichier PDF de fiches de suivi
    @param data une liste d'instances de Suivi
    @param tabou une liste de positions où on ne doit pas imprimer
       d'étiquette ; chaque position est donnée par un dictionnaire
       "row" : entier, "col": entier (numéros débutant à un)
    @return un pseudo-fichier, ouvert en lecteur, initialisé à son début
    """
    sheet = labels.Sheet(Suivi.specs, Suivi.draw_etiquette,
                         border=False)
    # dans le cas où le nombre de fiches de suivi est impair, on imprime
    # seulement à la deuxième position, afin de pouvoir placer une feuille
    # A5 comme première feuille à imprimer
    if len(data) %2 == 1:
        sheet.partial_page(1,[(1, 1)])
    for etiquette in data:
        sheet.add_label(etiquette)
    buffer = io.BytesIO()
    sheet.save(buffer)
    buffer.seek(0)
    return buffer

def make_carte_membre_verso_PDF(data):
    """
    Produit un fichier PDF de cartes de membre
    @param data une liste d'instances de CartesMembre
    @return un pseudo-fichier, ouvert en lecteur, initialisé à son début
    """
    sheet = labels.Sheet(CartesMembre.specs, CartesMembre.draw_etiquette,
                         border=False)
    for etiquette in data:
        sheet.add_label(etiquette)
    buffer = io.BytesIO()
    sheet.save(buffer)
    buffer.seek(0)
    return buffer

def demo():
    # démonstration !
    # création des feuilles à imprimer
    sheet = labels.Sheet(
        EtiquetteJeanBart.specs, EtiquetteJeanBart.draw_etiquette, border=True)

    # Ajout de quelques étiquettes
    for n in range(21365, 21370):
        e = EtiquetteJeanBart(
            n,                           # numéro pour le code-barre
            "Hispamundo Term G et STMG", # titre
            'ESP',                       # abrégé
            "14/03/2023",                # date
        )
        sheet.add_label(e)


    # Enregistrement du fichier et message
    sheet.save('essai.pdf')
    print("{0:d} étiquette(s) imprimées, sur {1:d} page(s).".format(
        sheet.label_count, sheet.page_count))
    return
    

if __name__ == "__main__":
    demo()
